"""Estimation of SPD matrices."""
import numpy as np
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.covariance import shrunk_covariance
from sklearn.metrics.pairwise import pairwise_kernels

from .spatialfilters import Xdawn
from .utils.covariance import (covariances, covariances_EP, cross_spectrum,
                               coherence, block_covariances)
from .utils import deprecated


def _nextpow2(i):
    """Find next power of 2."""
    n = 1
    while n < i:
        n *= 2
    return n


class Covariances(BaseEstimator, TransformerMixin):
    """Estimation of covariance matrices.

    Perform a simple covariance matrix estimation for each given input.

    Parameters
    ----------
    estimator : string, default='scm'
        Covariance matrix estimator, see
        :func:`pyriemann.utils.covariance.covariances`.
    **kwds : dict
        Any further parameters are passed directly to the covariance estimator.

    See Also
    --------
    ERPCovariances
    XdawnCovariances
    """

    def __init__(self, estimator='scm', **kwds):
        """Init."""
        self.estimator = estimator
        self.kwds = kwds

    def fit(self, X, y=None):
        """Fit.

        Do nothing. For compatibility purpose.

        Parameters
        ----------
        X : ndarray, shape (n_matrices, n_channels, n_times)
            Multi-channel time-series.
        y : None
            Not used, here for compatibility with sklearn API.

        Returns
        -------
        self : Covariances instance
            The Covariances instance.
        """
        return self

    def transform(self, X):
        """Estimate covariance matrices.

        Parameters
        ----------
        X : ndarray, shape (n_matrices, n_channels, n_times)
            Multi-channel time-series.

        Returns
        -------
        covmats : ndarray, shape (n_matrices, n_channels, n_channels)
            Covariance matrices.
        """
        covmats = covariances(X, estimator=self.estimator, **self.kwds)
        return covmats


class ERPCovariances(BaseEstimator, TransformerMixin):
    r"""Estimate special form covariance matrices for ERP.

    Estimation of special form covariance matrix dedicated to event-related
    potentials (ERP) processing.
    For each class, a prototyped response is obtained by average across trials:

    .. math::
        \mathbf{P} = \frac{1}{m} \sum_{i=1}^{m} \mathbf{X}_i

    and a super trial is built using the concatenation of :math:`\mathbf{P}`
    and the trial :math:`\mathbf{X}_i`:

    .. math::
        \mathbf{\tilde{X}}_i = \left[ \begin{array}{c} \mathbf{P} \\
        \mathbf{X}_i \end{array} \right]

    This super trial :math:`\mathbf{\tilde{X}}_i` will be used for covariance
    estimation.
    This allows to take into account the spatial structure of the signal, as
    described in [1]_.

    Parameters
    ----------
    classes : list of int | None, default=None
        List of classes to take into account for prototype estimation.
        If None, all classes will be accounted.
    estimator : string, default='scm'
        Covariance matrix estimator, see
        :func:`pyriemann.utils.covariance.covariances`.
    svd : int | None, default=None
        If not None, number of components of SVD used to reduce prototype
        responses.
    **kwds : dict
        Any further parameters are passed directly to the covariance estimator.

    Attributes
    ----------
    P_ : ndarray, shape (n_components, n_times)
        If fit, prototyped responses for each class, where `n_components` is
        equal to `n_classes x n_channels` if `svd` is None,
        and to `n_classes x min(svd, n_channels)` otherwise.

    See Also
    --------
    Covariances
    XdawnCovariances

    References
    ----------
    .. [1] `A Plug and Play P300 BCI Using Information Geometry
        <https://arxiv.org/abs/1409.0107>`_
        A. Barachant, M. Congedo. Research report, 2014.
    .. [2] `A New generation of Brain-Computer Interface Based on Riemannian
        Geometry
        <https://hal.archives-ouvertes.fr/hal-00879050>`_
        M. Congedo, A. Barachant, A. Andreev. Research report, 2013.
    .. [3] `Classification de potentiels evoques P300 par geometrie
        riemannienne pour les interfaces cerveau-machine EEG
        <https://hal.archives-ouvertes.fr/hal-00877447>`_
        A. Barachant, M. Congedo, G. van Veen, and C. Jutten, 24eme colloque
        GRETSI, 2013.
    """

    def __init__(self, classes=None, estimator='scm', svd=None, **kwds):
        """Init."""
        self.classes = classes
        self.estimator = estimator
        self.svd = svd
        self.kwds = kwds

    def fit(self, X, y):
        """Fit.

        Estimate the prototyped responses for each class.

        Parameters
        ----------
        X : ndarray, shape (n_matrices, n_channels, n_times)
            Multi-channel time-series.
        y : ndarray, shape (n_matrices,)
            Labels for each matrix.

        Returns
        -------
        self : ERPCovariances instance
            The ERPCovariances instance.
        """
        if self.svd is not None:
            if not isinstance(self.svd, int):
                raise TypeError('svd must be None or int')
        if self.classes is not None:
            classes = self.classes
        else:
            classes = np.unique(y)

        self.P_ = []
        for c in classes:
            # Prototyped response for each class
            P = np.mean(X[y == c], axis=0)

            # Apply svd if requested
            if self.svd is not None:
                U, _, _ = np.linalg.svd(P)
                P = U[:, 0:self.svd].T @ P

            self.P_.append(P)

        self.P_ = np.concatenate(self.P_, axis=0)
        return self

    def transform(self, X):
        """Estimate special form covariance matrices.

        Parameters
        ----------
        X : ndarray, shape (n_matrices, n_channels, n_times)
            Multi-channel time-series.

        Returns
        -------
        covmats : ndarray, shape (n_matrices, n_components, n_components)
            Covariance matrices for ERP, where the size of matrices
            `n_components` is equal to `(1 + n_classes) x n_channels` if `svd`
            is None, and to `n_channels + n_classes x min(svd, n_channels)`
            otherwise.
        """
        covmats = covariances_EP(
            X,
            self.P_,
            estimator=self.estimator,
            **self.kwds
        )
        return covmats


class XdawnCovariances(BaseEstimator, TransformerMixin):
    """Estimate special form covariance matrices for ERP combined with Xdawn.

    Estimation of special form covariance matrix dedicated to ERP processing
    combined with `Xdawn` spatial filtering.
    This is similar to :class:`pyriemann.estimation.ERPCovariances` but data
    are spatially filtered with :class:`pyriemann.spatialfilters.Xdawn`.
    A complete description of the method is available in [1]_.

    The advantage of this estimation is to reduce dimensionality of the
    covariance matrices supervisely.

    Parameters
    ----------
    nfilter : int, default=4
        Number of Xdawn filters per class.
    applyfilters : bool, default=True
        If set to true, spatial filter are applied to the prototypes and the
        signals. When set to False, filters are applied only to the ERP
        prototypes allowing for a better generalization across subject and
        session at the expense of dimensionality increase. In that case, the
        estimation is similar to :class:`pyriemann.estimation.ERPCovariances`
        with `svd=nfilter` but with more compact prototype reduction.
    classes : list of int | None, default=None
        list of classes to take into account for prototype estimation.
        If None, all classes will be accounted.
    estimator : string, default='scm'
        Covariance matrix estimator, see
        :func:`pyriemann.utils.covariance.covariances`.
    xdawn_estimator : string, default='scm'
        Covariance matrix estimator for `Xdawn` spatial filtering.
        Should be regularized using 'lwf' or 'oas', see
        :func:`pyriemann.utils.covariance.covariances`.
    baseline_cov : array, shape (n_channels, n_channels) | None, default=None
        Baseline covariance for `Xdawn` spatial filtering,
        see :class:`pyriemann.spatialfilters.Xdawn`.
    **kwds : dict
        Any further parameters are passed directly to the covariance estimator.

    Attributes
    ----------
    P_ : ndarray, shape (n_classes x min(n_channels, n_filters), n_times)
        If fit, the evoked response for each event type, concatenated.

    See Also
    --------
    ERPCovariances
    Xdawn

    References
    ----------
    .. [1] `MEG decoding using Riemannian Geometry and
        Unsupervised classification
        <https://citeseerx.ist.psu.edu/viewdoc/summary?doi=10.1.1.713.5131>`_
        A. Barachant. Technical report with the solution of the DecMeg 2014
        challenge.
    """

    def __init__(self,
                 nfilter=4,
                 applyfilters=True,
                 classes=None,
                 estimator='scm',
                 xdawn_estimator='scm',
                 baseline_cov=None,
                 **kwds):
        """Init."""
        self.applyfilters = applyfilters
        self.estimator = estimator
        self.xdawn_estimator = xdawn_estimator
        self.classes = classes
        self.nfilter = nfilter
        self.baseline_cov = baseline_cov
        self.kwds = kwds

    def fit(self, X, y):
        """Fit.

        Estimate spatial filters and prototyped response for each classes.

        Parameters
        ----------
        X : ndarray, shape (n_matrices, n_channels, n_times)
            Multi-channel time-series.
        y : ndarray, shape (n_matrices,)
            Labels for each matrix.

        Returns
        -------
        self : XdawnCovariances instance
            The XdawnCovariances instance.
        """
        self.Xd_ = Xdawn(
            nfilter=self.nfilter,
            classes=self.classes,
            estimator=self.xdawn_estimator,
            baseline_cov=self.baseline_cov,
        )
        self.Xd_.fit(X, y)
        self.P_ = self.Xd_.evokeds_
        return self

    def transform(self, X):
        """Estimate Xdawn covariance matrices.

        Parameters
        ----------
        X : ndarray, shape (n_matrices, n_channels, n_times)
            Multi-channel time-series.

        Returns
        -------
        covmats : ndarray, shape (n_matrices, n_components, n_components)
            Covariance matrices filtered by Xdawn, where n_components is equal
            to `2 x n_classes x min(n_channels, nfilter)` if `applyfilters` is
            True, and to `n_channels + n_classes x min(n_channels, nfilter)`
            otherwise.
        """
        if self.applyfilters:
            X = self.Xd_.transform(X)

        covmats = covariances_EP(
            X,
            self.P_,
            estimator=self.estimator,
            **self.kwds
        )
        return covmats


class BlockCovariances(BaseEstimator, TransformerMixin):
    """Estimation of block covariance matrices.

    Perform a block covariance estimation for each given matrix. The
    resulting matrices are block diagonal matrices.

    The blocks on the diagonal are calculated as individual covariance
    matrices for a subset of channels using the given the estimator.
    Varying block sized possible by passing a list to allow incorporation
    of different modalities with different number of channels (e.g. EEG,
    ECoG, LFP, EMG) with their own respective covariance matrices.

    Parameters
    ----------
    block_size : int | list of int
        Sizes of individual blocks given as int for same-size block, or list
        for varying block sizes.
    estimator : string, default='scm'
        Covariance matrix estimator, see
        :func:`pyriemann.utils.covariance.covariances`.
    **kwds : dict
        Any further parameters are passed directly to the covariance estimator.

    Notes
    -----
    .. versionadded:: 0.3

    See Also
    --------
    Covariances
    """

    def __init__(self, block_size, estimator='scm', **kwds):
        """Init."""
        self.estimator = estimator
        self.block_size = block_size
        self.kwds = kwds

    def fit(self, X, y=None):
        """Fit.

        Do nothing. For compatibility purpose.

        Parameters
        ----------
        X : ndarray, shape (n_matrices, n_channels, n_times)
            Multi-channel time-series.
        y : None
            Not used, here for compatibility with sklearn API.

        Returns
        -------
        self : BlockCovariances instance
            The BlockCovariances instance.
        """
        return self

    def transform(self, X):
        """Estimate block covariance matrices.

        Parameters
        ----------
        X : ndarray, shape (n_matrices, n_channels, n_times)
            Multi-channel time-series.

        Returns
        -------
        covmats : ndarray, shape (n_matrices, n_channels, n_channels)
            Covariance matrices.
        """
        n_matrices, n_channels, n_times = X.shape

        if isinstance(self.block_size, int):
            n_blocks = n_channels // self.block_size
            blocks = [self.block_size for b in range(n_blocks)]

        elif isinstance(self.block_size, (list, np.ndarray)):
            blocks = self.block_size

        else:
            raise ValueError("Parameter block_size must be int or list.")

        return block_covariances(X, blocks, self.estimator, **self.kwds)


###############################################################################


class CrossSpectra(BaseEstimator, TransformerMixin):
    """Estimation of cross-spectral matrices.

    Complex cross-spectral matrices are HPD matrices estimated as the spectrum
    covariance in the frequency domain [1]_. It returns a 4-d array with a
    cross-spectral matrix for each input and in each frequency bin of the
    Fourier transform.

    Parameters
    ----------
    window : int, default=128
        The length of the FFT window used for spectral estimation.
    overlap : float, default=0.75
        The percentage of overlap between window.
    fmin : float | None, default=None
        The minimal frequency to be returned.
    fmax : float | None, default=None
        The maximal frequency to be returned.
    fs : float | None, default=None
        The sampling frequency of the signal.

    Attributes
    ----------
    freqs_ : ndarray, shape (n_freqs,)
        If transformed, the frequencies associated to cross-spectra.
        None if ``fs`` is None.

    Notes
    -----
    .. versionadded:: 0.6

    See Also
    --------
    CoSpectra
    Coherences

    References
    ----------
    .. [1] https://en.wikipedia.org/wiki/Cross-spectrum
    """

    def __init__(self, window=128, overlap=0.75, fmin=None, fmax=None,
                 fs=None):
        """Init."""
        self.window = _nextpow2(window)
        self.overlap = overlap
        self.fmin = fmin
        self.fmax = fmax
        self.fs = fs

    def fit(self, X, y=None):
        """Fit.

        Do nothing. For compatibility purpose.

        Parameters
        ----------
        X : ndarray, shape (n_matrices, n_channels, n_times)
            Multi-channel time-series.
        y : None
            Not used, here for compatibility with sklearn API.

        Returns
        -------
        self : CrossSpectra instance
            The CrossSpectra instance.
        """
        return self

    def transform(self, X):
        """Estimate cross-spectral matrices.

        Parameters
        ----------
        X : ndarray, shape (n_matrices, n_channels, n_times)
            Multi-channel time-series.

        Returns
        -------
        X_new : ndarray, shape (n_matrices, n_channels, n_channels, n_freqs)
            Cross-spectral matrices for each input and for each frequency bin.
        """
        X_new = []

        for i in range(len(X)):
            S, freqs = cross_spectrum(
                X[i],
                window=self.window,
                overlap=self.overlap,
                fmin=self.fmin,
                fmax=self.fmax,
                fs=self.fs)
            X_new.append(S)
        self.freqs_ = freqs

        return np.array(X_new)


class CoSpectra(CrossSpectra):
    """Estimation of co-spectral matrices.

    Co-spectral matrices are SPD matrices estimated as the real part of the
    :class:`pyriemann.estimation.CrossSpectra`. It returns a 4-d array with a
    co-spectral matrix for each input and in each frequency bin of the
    Fourier transform.

    Parameters
    ----------
    window : int, default=128
        The length of the FFT window used for spectral estimation.
    overlap : float, default=0.75
        The percentage of overlap between window.
    fmin : float | None, default=None
        The minimal frequency to be returned.
    fmax : float | None, default=None
        The maximal frequency to be returned.
    fs : float | None, default=None
        The sampling frequency of the signal.

    Attributes
    ----------
    freqs_ : ndarray, shape (n_freqs,)
        If transformed, the frequencies associated to cospectra.
        None if ``fs`` is None.

    See Also
    --------
    CrossSpectra
    Coherences
    """

    def transform(self, X):
        """Estimate co-spectral matrices.

        Parameters
        ----------
        X : ndarray, shape (n_matrices, n_channels, n_times)
            Multi-channel time-series.

        Returns
        -------
        X_new : ndarray, shape (n_matrices, n_channels, n_channels, n_freqs)
            Co-spectral matrices for each input and for each frequency bin.
        """
        X_new = super().transform(X)
        return X_new.real


@deprecated(
    "CospCovariances is deprecated and will be removed in 0.8.0; "
    "please use CoSpectra."
)
class CospCovariances(CoSpectra):
    pass


class Coherences(CospCovariances):
    """Estimation of squared coherence matrices.

    Squared coherence matrices estimation [1]_. This method will return a 4-d
    array with a squared coherence matrix estimation for each input and in
    each frequency bin of the FFT.

    Parameters
    ----------
    window : int, default=128
        The length of the FFT window used for spectral estimation.
    overlap : float, default=0.75
        The percentage of overlap between window.
    fmin : float | None, default=None
        the minimal frequency to be returned.
    fmax : float | None, default=None
        The maximal frequency to be returned.
    fs : float | None, default=None
        The sampling frequency of the signal.
    coh : {'ordinary', 'instantaneous', 'lagged', 'imaginary'}, \
            default='ordinary'
        The coherence type:

        * 'ordinary' for the ordinary coherence, defined in Eq.(22) of [1]_;
          this normalization of cross-spectral matrices captures both in-phase
          and out-of-phase correlations. However it is inflated by the
          artificial in-phase (zero-lag) correlation engendered by volume
          conduction.
        * 'instantaneous' for the instantaneous coherence, Eq.(26) of [1]_,
          capturing only in-phase correlation.
        * 'lagged' for the lagged-coherence, Eq.(28) of [1]_, capturing only
          out-of-phase correlation (not defined for DC and Nyquist bins).
        * 'imaginary' for the imaginary coherence [2]_, Eq.(0.16) of [3]_,
          capturing out-of-phase correlation but still affected by in-phase
          correlation.

    Attributes
    ----------
    freqs_ : ndarray, shape (n_freqs,)
        If transformed, the frequencies associated to cospectra.
        None if ``fs`` is None.

    Notes
    -----
    .. versionadded:: 0.3

    See Also
    --------
    CrossSpectra
    TimeDelayCovariances

    References
    ----------
    .. [1] `Instantaneous and lagged measurements of linear
        and nonlinear dependence between groups of multivariate time series:
        frequency decomposition
        <https://arxiv.org/ftp/arxiv/papers/0711/0711.1455.pdf>`_
        R. Pascual-Marqui. Technical report, 2007.
    .. [2] `Identifying true brain interaction from EEG data using the
        imaginary part of coherency
        <https://doi.org/10.1016/j.clinph.2004.04.029>`_
        G. Nolte, O. Bai, L. Wheaton, Z. Mari, S. Vorbach, M. Hallett.
        Clinical Neurophysioly, Volume 115, Issue 10, October 2004,
        Pages 2292-2307
    .. [3] `Non-Parametric Synchronization Measures used in EEG
        and MEG
        <https://hal.archives-ouvertes.fr/hal-01868538v2>`_
        M. Congedo. Technical Report, 2018.
    """

    def __init__(self, window=128, overlap=0.75, fmin=None, fmax=None,
                 fs=None, coh='ordinary'):
        """Init."""
        self.window = _nextpow2(window)
        self.overlap = overlap
        self.fmin = fmin
        self.fmax = fmax
        self.fs = fs
        self.coh = coh

    def transform(self, X):
        """Estimate the squared coherences matrices.

        Parameters
        ----------
        X : ndarray, shape (n_matrices, n_channels, n_times)
            Multi-channel time-series.

        Returns
        -------
        covmats : ndarray, shape (n_matrices, n_channels, n_channels, n_freqs)
            Squared coherence matrices for each input and for each frequency
            bin.
        """
        out = []

        for i in range(len(X)):
            S, freqs = coherence(
                X[i],
                window=self.window,
                overlap=self.overlap,
                fmin=self.fmin,
                fmax=self.fmax,
                fs=self.fs,
                coh=self.coh)
            out.append(S)
        self.freqs_ = freqs

        return np.array(out)


class TimeDelayCovariances(BaseEstimator, TransformerMixin):
    """Estimation of covariance matrices with time delay matrices.

    Time delay covariance matrices are useful to catch spectral dynamics of
    the signal, similarly to the CSSP method [1]_. It is done by concatenating
    time delayed version of the signal before covariance estimation.

    Parameters
    ----------
    delays : int | list of int, default=4
        The delays to apply for the Hankel matrices. If `int`, it use a range
        of delays up to the given value. A list of int can be given.
    estimator : string, default='scm'
        Covariance matrix estimator, see
        :func:`pyriemann.utils.covariance.covariances`.
    **kwds : dict
        Any further parameters are passed directly to the covariance estimator.

    Attributes
    ----------
    Xtd_ : ndarray, shape (n_matrices, n_channels x n_delays, n_times)
        Time delay multi-channel time-series, where `n_delays` is equal to:
        `delays` when it is a int, and `1 + len(delays)` when it is a list.

    See Also
    --------
    Covariances
    ERPCovariances
    CospCovariances

    References
    ----------
    .. [1] `Spatio-spectral filters for improving the classification of single
        trial EEG
        <http://doc.ml.tu-berlin.de/bbci/publications/LemBlaCurMue05.pdf>`_
        S. Lemm, B. Blankertz, B. Curio, K-R. Muller. IEEE Transactions on
        Biomedical Engineering 52(9), 1541-1548, 2005.
    """

    def __init__(self, delays=4, estimator='scm', **kwds):
        """Init."""
        self.delays = delays
        self.estimator = estimator
        self.kwds = kwds

    def fit(self, X, y=None):
        """Fit.

        Do nothing. For compatibility purpose.

        Parameters
        ----------
        X : ndarray, shape (n_matrices, n_channels, n_times)
            Multi-channel time-series.
        y : None
            Not used, here for compatibility with sklearn API.

        Returns
        -------
        self : TimeDelayCovariances instance
            The TimeDelayCovariances instance.
        """
        return self

    def transform(self, X):
        """Estimate the time delay covariance matrices.

        Parameters
        ----------
        X : ndarray, shape (n_matrices, n_channels, n_times)
            Multi-channel time-series.

        Returns
        -------
        covmats : ndarray, shape (n_matrices, n_channels x n_delays, \
                n_channels x n_delays)
            Time delay covariance matrices, where `n_delays` is equal to:
            `delays` when it is a int, and `1 + len(delays)` when it is a list.
        """

        if isinstance(self.delays, int):
            delays = range(1, self.delays)
        elif isinstance(self.delays, list):
            delays = self.delays
        else:
            raise ValueError('delays must be an integer or a list')

        Xtd = [X]
        for d in delays:
            Xtd.append(np.roll(X, d, axis=-1))
        self.Xtd_ = np.concatenate(Xtd, axis=-2)

        covmats = covariances(self.Xtd_, estimator=self.estimator, **self.kwds)
        return covmats


@deprecated(
    "HankelCovariances is deprecated and will be removed in 0.8.0; "
    "please use TimeDelayCovariances."
)
class HankelCovariances(TimeDelayCovariances):
    pass


###############################################################################


class Kernels(BaseEstimator, TransformerMixin):
    r"""Estimation of kernel matrices between channels of time series.

    Perform a kernel matrix estimation for each given time series, evaluating a
    kernel function between each pair of channels (rather than between pairs of
    time samples) and allowing to extract nonlinear channel relationship [1]_.

    For an input time series :math:`X \in \mathbb{R}^{c \times t}`, composed of
    :math:`c` channels and :math:`t` time samples, kernel function
    :math:`\kappa()` is computed between channels :math:`i` and :math:`j`:

    .. math::
        K_{i,j} = \kappa \left( X[i], X[j] \right)

    Linear kernel is related to :class:`pyriemann.estimation.Covariances` [1]_,
    but this class allows to generalize to nonlinear relationships.

    Parameters
    ----------
    metric : string, default='linear'
        The metric to use when computing kernel function between channels [2]_:
        'linear', 'poly', 'polynomial', 'rbf', 'laplacian', 'cosine'.
    n_jobs : int, default=None
        The number of jobs to use for the computation [2]_. This works by
        breaking down the pairwise matrix into n_jobs even slices and computing
        them in parallel.
    **kwds : dict
        Any further parameters are passed directly to the kernel function [2]_.

    See Also
    --------
    Covariances

    Notes
    -----
    .. versionadded:: 0.4

    References
    ----------
    .. [1] `Beyond Covariance: Feature Representation with Nonlinear Kernel
        Matrices
        <https://www.cv-foundation.org/openaccess/content_iccv_2015/papers/Wang_Beyond_Covariance_Feature_ICCV_2015_paper.pdf>`_
        L. Wang, J. Zhang, L. Zhou, C. Tang, W Li. ICCV, 2015.
    .. [2]
        https://scikit-learn.org/stable/modules/generated/sklearn.metrics.pairwise.pairwise_kernels.html
    """  # noqa

    def __init__(self, metric='linear', n_jobs=None, **kwds):
        """Init."""
        self.metric = metric
        self.n_jobs = n_jobs
        self.kwds = kwds

    def fit(self, X, y=None):
        """Fit.

        Do nothing. For compatibility purpose.

        Parameters
        ----------
        X : ndarray, shape (n_matrices, n_channels, n_times)
            Multi-channel time-series.
        y : None
            Not used, here for compatibility with sklearn API.

        Returns
        -------
        self : Kernels instance
            The Kernels instance.
        """
        return self

    def transform(self, X):
        """Estimate kernel matrices from time series.

        Parameters
        ----------
        X : ndarray, shape (n_matrices, n_channels, n_times)
            Multi-channel time-series.

        Returns
        -------
        K : ndarray, shape (n_matrices, n_channels, n_channels)
            Kernel matrices.
        """
        if self.metric not in [
                'linear', 'poly', 'polynomial', 'rbf', 'laplacian', 'cosine'
        ]:
            raise TypeError('Unsupported metric for kernel estimation.')

        K = [
            pairwise_kernels(
                x,
                None,
                metric=self.metric,
                n_jobs=self.n_jobs,
                **self.kwds
            ) for x in X
        ]

        return np.asarray(K)


###############################################################################


class Shrinkage(BaseEstimator, TransformerMixin):
    """Regularization of SPD matrices by shrinkage.

    This transformer applies a shrinkage regularization to any SPD matrix.
    It directly uses the `shrunk_covariance` function from scikit-learn [1]_,
    applied on each input.

    Parameters
    ----------
    shrinkage : float, default=0.1
        Coefficient in the convex combination used for the computation of the
        shrunk estimate. Must be between 0 and 1.

    Notes
    -----
    .. versionadded:: 0.2.5

    References
    ----------
    .. [1] https://scikit-learn.org/stable/modules/generated/sklearn.covariance.shrunk_covariance.html
    """  # noqa

    def __init__(self, shrinkage=0.1):
        """Init."""
        self.shrinkage = shrinkage

    def fit(self, X, y=None):
        """Fit.

        Do nothing. For compatibility purpose.

        Parameters
        ----------
        X : ndarray, shape (n_matrices, n_channels, n_channels)
            Set of SPD matrices.
        y : None
            Not used, here for compatibility with sklearn API.

        Returns
        -------
        self : Shrinkage instance
            The Shrinkage instance.
        """
        return self

    def transform(self, X):
        """Shrink and return the SPD matrices.

        Parameters
        ----------
        X : ndarray, shape (n_matrices, n_channels, n_channels)
            Set of SPD matrices.

        Returns
        -------
        covmats : ndarray, shape (n_matrices, n_channels, n_channels)
            Set of shrunk SPD matrices.
        """

        covmats = np.zeros_like(X)

        for ii, x in enumerate(X):
            covmats[ii] = shrunk_covariance(x, self.shrinkage)

        return covmats



"""Base functions for SPD/HPD matrices."""

import numpy as np
from numpy.core.numerictypes import typecodes
from .test import is_pos_def


def _matrix_operator(C, operator):
    """Matrix function."""
    if not isinstance(C, np.ndarray) or C.ndim < 2:
        raise ValueError("Input must be at least a 2D ndarray")
    if C.dtype.char in typecodes['AllFloat'] and (
            np.isinf(C).any() or np.isnan(C).any()):
        raise ValueError(
            "Matrices must be positive definite. Add "
            "regularization to avoid this error.")
    eigvals, eigvecs = np.linalg.eigh(C)
    eigvals = operator(eigvals)
    if C.ndim >= 3:
        eigvals = np.expand_dims(eigvals, -2)
    D = (eigvecs * eigvals) @ np.swapaxes(eigvecs.conj(), -2, -1)
    return D


def expm(C):
    r"""Exponential of SPD/HPD matrices.

    The symmetric matrix exponential of a SPD/HPD matrix
    :math:`\mathbf{C}` is defined by:

    .. math::
        \mathbf{D} = \mathbf{V} \exp{(\mathbf{\Lambda})} \mathbf{V}^H

    where :math:`\mathbf{\Lambda}` is the diagonal matrix of eigenvalues
    and :math:`\mathbf{V}` the eigenvectors of :math:`\mathbf{C}`.

    Parameters
    ----------
    C : ndarray, shape (..., n, n)
        SPD/HPD matrices, at least 2D ndarray.

    Returns
    -------
    D : ndarray, shape (..., n, n)
        Matrix exponential of C.
    """
    return _matrix_operator(C, np.exp)


def invsqrtm(C):
    r"""Inverse square root of SPD/HPD matrices.

    The symmetric matrix inverse square root of a SPD/HPD matrix
    :math:`\mathbf{C}` is defined by:

    .. math::
        \mathbf{D} =
        \mathbf{V} \left( \mathbf{\Lambda} \right)^{-1/2} \mathbf{V}^H

    where :math:`\mathbf{\Lambda}` is the diagonal matrix of eigenvalues
    and :math:`\mathbf{V}` the eigenvectors of :math:`\mathbf{C}`.

    Parameters
    ----------
    C : ndarray, shape (..., n, n)
        SPD/HPD matrices, at least 2D ndarray.

    Returns
    -------
    D : ndarray, shape (..., n, n)
        Matrix inverse square root of C.
    """
    def isqrt(x): return 1. / np.sqrt(x)
    return _matrix_operator(C, isqrt)


def logm(C):
    r"""Logarithm of SPD/HPD matrices.

    The symmetric matrix logarithm of a SPD/HPD matrix
    :math:`\mathbf{C}` is defined by:

    .. math::
        \mathbf{D} = \mathbf{V} \log{(\mathbf{\Lambda})} \mathbf{V}^H

    where :math:`\mathbf{\Lambda}` is the diagonal matrix of eigenvalues
    and :math:`\mathbf{V}` the eigenvectors of :math:`\mathbf{C}`.

    Parameters
    ----------
    C : ndarray, shape (..., n, n)
        SPD/HPD matrices, at least 2D ndarray.

    Returns
    -------
    D : ndarray, shape (..., n, n)
        Matrix logarithm of C.
    """
    return _matrix_operator(C, np.log)


def powm(C, alpha):
    r"""Power of SPD/HPD matrices.

    The symmetric matrix power :math:`\alpha` of a SPD/HPD matrix
    :math:`\mathbf{C}` is defined by:

    .. math::
        \mathbf{D} =
        \mathbf{V} \left( \mathbf{\Lambda} \right)^{\alpha} \mathbf{V}^H

    where :math:`\mathbf{\Lambda}` is the diagonal matrix of eigenvalues
    and :math:`\mathbf{V}` the eigenvectors of :math:`\mathbf{C}`.

    Parameters
    ----------
    C : ndarray, shape (..., n, n)
        SPD/HPD matrices, at least 2D ndarray.
    alpha : float
        The power to apply.

    Returns
    -------
    D : ndarray, shape (..., n, n)
        Matrix power of C.
    """
    def power(x): return x**alpha
    return _matrix_operator(C, power)


def sqrtm(C):
    r"""Square root of SPD/HPD matrices.

    The symmetric matrix square root of a SPD/HPD matrix
    :math:`\mathbf{C}` is defined by:

    .. math::
        \mathbf{D} =
        \mathbf{V} \left( \mathbf{\Lambda} \right)^{1/2} \mathbf{V}^H

    where :math:`\mathbf{\Lambda}` is the diagonal matrix of eigenvalues
    and :math:`\mathbf{V}` the eigenvectors of :math:`\mathbf{C}`.

    Parameters
    ----------
    C : ndarray, shape (..., n, n)
        SPD/HPD matrices, at least 2D ndarray.

    Returns
    -------
    D : ndarray, shape (..., n, n)
        Matrix square root of C.
    """
    return _matrix_operator(C, np.sqrt)


###############################################################################


def _nearest_sym_pos_def(S, reg=1e-6):
    """Find the nearest SPD matrix.

    Parameters
    ----------
    S : ndarray, shape (n, n)
        Square matrix.
    reg : float
        Regularization parameter.

    Returns
    -------
    P : ndarray, shape (n, n)
        Nearest SPD matrix.
    """
    A = (S + S.T) / 2
    _, s, V = np.linalg.svd(A)
    H = V.T @ np.diag(s) @ V
    B = (A + H) / 2
    P = (B + B.T) / 2

    if is_pos_def(P):
        # Regularize if already PD
        ei, ev = np.linalg.eigh(P)
        if np.min(ei) / np.max(ei) < reg:
            P = ev @ np.diag(ei + reg) @ ev.T
        return P

    spacing = np.spacing(np.linalg.norm(A))
    I = np.eye(S.shape[0])  # noqa
    k = 1
    while not is_pos_def(P, fast_mode=False):
        mineig = np.min(np.real(np.linalg.eigvals(P)))
        P += I * (-mineig * k ** 2 + spacing)
        k += 1

    # Regularize
    ei, ev = np.linalg.eigh(P)
    if np.min(ei) / np.max(ei) < reg:
        P = ev @ np.diag(ei + reg) @ ev.T
    return P


def nearest_sym_pos_def(X, reg=1e-6):
    """Find the nearest SPD matrices.

    A NumPy port of John D'Errico's `nearestSPD` MATLAB code [1]_,
    which credits [2]_.

    Parameters
    ----------
    X : ndarray, shape (..., n, n)
        Square matrices, at least 2D ndarray.
    reg : float
        Regularization parameter.

    Returns
    -------
    P : ndarray, shape (..., n, n)
        Nearest SPD matrices.

    Notes
    -----
    .. versionadded:: 0.4

    References
    ----------
    .. [1] `nearestSPD
        <https://www.mathworks.com/matlabcentral/fileexchange/42885-nearestspd>`_
        J. D'Errico, MATLAB Central File Exchange
    .. [2] `Computing a nearest symmetric positive semidefinite matrix
        <https://www.sciencedirect.com/science/article/pii/0024379588902236>`_
        N.J. Higham, Linear Algebra and its Applications, vol 103, 1988
    """
    return np.array([_nearest_sym_pos_def(x, reg) for x in X])



"""Medians of SPD/HPD matrices."""

import warnings
import numpy as np

from .base import sqrtm, invsqrtm, logm, expm
from .distance import distance
from .mean import mean_euclid
from .utils import check_weights


def median_euclid(X, *, tol=10e-6, maxiter=50, init=None, weights=None):
    r"""Euclidean geometric median of matrices.

    The Euclidean geometric median minimizes the sum of Euclidean distances
    :math:`d_E` to all matrices [1]_ [2]_:

    .. math::
        \arg \min_{\mathbf{M}} \sum_i w_i \ d_E (\mathbf{M}, \mathbf{X}_i)

    Geometric median is different from the marginal median provided by NumPy
    [3]_.

    Parameters
    ----------
    X : ndarray, shape (n_matrices, n, m)
        Set of matrices.
    tol : float, default=10e-6
        The tolerance to stop the iterative algorithm.
    maxiter : int, default=50
        The maximum number of iterations.
    init : None | ndarray, shape (n, m), default=None
        A matrix used to initialize the iterative algorithm.
        If None, the weighted Euclidean mean is used.
    weights : None | ndarray, shape (n_matrices,), default=None
        Weights for each matrix. If None, it uses equal weights.

    Returns
    -------
    M : ndarray, shape (n, m)
        Euclidean geometric median.

    Notes
    -----
    .. versionadded:: 0.4

    References
    ----------
    .. [1] `Sur le point pour lequel la somme des distances de n points donnés
        est minimum
        <https://www.jstage.jst.go.jp/article/tmj1911/43/0/43_0_355/_pdf>`_
        E Weiszfeld. Tohoku Mathematical Journal, 1937, 43, pp. 355-386.
    .. [2] `The multivariate L1-median and associated data depth
        <https://www.pnas.org/doi/pdf/10.1073/pnas.97.4.1423>`_
        Y Vardi and C-H Zhan. Proceedings of the National Academy of Sciences,
        2000, vol. 97, no 4, p. 1423-1426
    .. [3] https://numpy.org/doc/stable/reference/generated/numpy.median.html
    """
    n_matrices, _, _ = X.shape
    weights = check_weights(weights, n_matrices)
    if init is None:
        M = mean_euclid(X, sample_weight=weights)
    else:
        M = init

    for _ in range(maxiter):
        dists = distance(X, M, metric='euclid')[:, 0]
        is_zero = (dists == 0)

        w = weights[~is_zero] / dists[~is_zero]
        Mnew = mean_euclid(X[~is_zero], sample_weight=w)  # Eq(2.4) of [2]

        n_zeros = np.sum(is_zero)
        if n_zeros > 0:
            R = np.einsum('a,abc->bc', w, X[~is_zero] - M)  # Eq(2.7)
            r = np.linalg.norm(R, ord='fro')
            rinv = 0 if r == 0 else np.mean(weights[is_zero]) / r
            Mnew = max(0, 1 - rinv) * Mnew + min(1, rinv) * M  # Eq(2.6)

        crit = np.linalg.norm(Mnew - M, ord='fro')
        M = Mnew
        if crit <= tol:
            break
    else:
        warnings.warn("Convergence not reached")

    return M


def median_riemann(X, *, tol=10e-6, maxiter=50, init=None, weights=None,
                   step_size=1):
    r"""Affine-invariant Riemannian geometric median of SPD/HPD matrices.

    The affine-invariant Riemannian geometric median minimizes the sum of
    affine-invariant Riemannian distances :math:`d_R` to all SPD/HPD matrices
    [1]_:

    .. math::
        \arg \min_{\mathbf{M}} \sum_i w_i \ d_R (\mathbf{M}, \mathbf{X}_i)

    Parameters
    ----------
    X : ndarray, shape (n_matrices, n, n)
        Set of SPD/HPD matrices.
    tol : float, default=10e-6
        The tolerance to stop the gradient descent.
    maxiter : int, default=50
        The maximum number of iterations.
    init : None | ndarray, shape (n, n), default=None
        A SPD/HPD matrix used to initialize the gradient descent.
        If None, the weighted Euclidean mean is used.
    weights : None | ndarray, shape (n_matrices,), default=None
        Weights for each matrix. If None, it uses equal weights.
    step_size : float, default=1.0
        The step size of the gradient descent, in (0,2].

    Returns
    -------
    M : ndarray, shape (n, n)
        Affine-invariant Riemannian geometric median.

    Notes
    -----
    .. versionadded:: 0.4

    References
    ----------
    .. [1] `The geometric median on Riemannian manifolds with application to
        robust atlas estimation
        <https://www.ncbi.nlm.nih.gov/pmc/articles/PMC2735114/>`_
        PT. Fletcher, S. Venkatasubramanian S and S. Joshi.
        NeuroImage, 2009, 45(1), S143-S152
    .. [2] `Riemannian median, geometry of covariance matrices and radar target
        detection
        <https://ieeexplore.ieee.org/abstract/document/5615027>`_
        L Yang, M Arnaudon and F Barbaresco. 7th European Radar Conference,
        2010, pp. 415-418
    """
    if not 0 < step_size <= 2:
        raise ValueError(
            f"Value step_size must be included in (0, 2] (Got {step_size})")
    n_matrices, _, _ = X.shape
    weights = check_weights(weights, n_matrices)
    if init is None:
        M = mean_euclid(X, sample_weight=weights)
    else:
        M = init

    for _ in range(maxiter):
        dists = distance(X, M, metric='riemann')[:, 0]
        is_zero = (dists == 0)
        w = weights[~is_zero] / dists[~is_zero]

        # Eq(11) of [1]
        M12, Mm12 = sqrtm(M), invsqrtm(M)
        tangvecs = logm(Mm12 @ X[~is_zero] @ Mm12)
        J = np.einsum('a,abc->bc', w / np.sum(w), tangvecs)
        M = M12 @ expm(step_size * J) @ M12

        crit = np.linalg.norm(J, ord='fro')
        if crit <= tol:
            break
    else:
        warnings.warn("Convergence not reached")

    return M



import pickle

import numpy as np
from numpy.testing import assert_array_equal
import pytest

from conftest import get_distances, get_means, get_metrics
from pyriemann.regression import (SVR, KNearestNeighborRegressor)
from pyriemann.utils.kernel import kernel
from pyriemann.utils.mean import mean_covariance

regs = [SVR, KNearestNeighborRegressor]


@pytest.mark.parametrize("regres", regs)
class RegressorTestCase:
    def test_regression(self, regres, get_mats, get_targets):
        n_matrices, n_channels = 6, 3
        covmats = get_mats(n_matrices, n_channels, "spd")
        targets = get_targets(n_matrices)

        self.reg_predict(regres, covmats, targets)
        self.reg_fit_independence(regres, covmats, targets)
        self.reg_score(regres, covmats, targets)


class TestRegressor(RegressorTestCase):
    def reg_predict(self, regres, covmats, targets):
        clf = regres()
        clf.fit(covmats, targets)
        predicted = clf.predict(covmats)
        assert predicted.shape == (len(targets),)

    def reg_fit_independence(self, regres, covmats, targets):
        clf = regres()
        clf.fit(covmats, targets).predict(covmats)
        # retraining with different size should erase previous fit
        new_covmats = covmats[:, :-1, :-1]
        clf.fit(new_covmats, targets).predict(new_covmats)

    def reg_score(self, regres, covmats, targets):
        clf = regres()
        clf.fit(covmats, targets)
        clf.score(covmats, targets)


@pytest.mark.parametrize("regres", [KNearestNeighborRegressor])
@pytest.mark.parametrize("mean", ["faulty", 42])
@pytest.mark.parametrize("dist", ["not_real", 27])
def test_metric_dict_error(regres, mean, dist, get_mats, get_targets):
    n_matrices, n_channels = 6, 3
    targets = get_targets(n_matrices)
    covmats = get_mats(n_matrices, n_channels, "spd")
    with pytest.raises(ValueError):
        clf = regres(metric={"mean": mean, "distance": dist})
        clf.fit(covmats, targets).predict(covmats)


@pytest.mark.parametrize("regres", [KNearestNeighborRegressor])
@pytest.mark.parametrize("mean", get_means())
@pytest.mark.parametrize("dist", get_distances())
def test_metric_dist(regres, mean, dist, get_mats, get_targets):
    n_matrices, n_channels = 4, 3
    targets = get_targets(n_matrices)
    covmats = get_mats(n_matrices, n_channels, "spd")
    clf = regres(metric={"mean": mean, "distance": dist})
    clf.fit(covmats, targets).predict(covmats)


@pytest.mark.parametrize("regres", regs)
@pytest.mark.parametrize("metric", [42, "faulty", {"foo": "bar"}])
def test_metric_wrong_keys(regres, metric, get_mats, get_targets):
    n_matrices, n_channels = 6, 3
    targets = get_targets(n_matrices)
    covmats = get_mats(n_matrices, n_channels, "spd")
    with pytest.raises((TypeError, KeyError, ValueError)):
        clf = regres(metric=metric)
        clf.fit(covmats, targets).predict(covmats)


@pytest.mark.parametrize("regres", regs)
@pytest.mark.parametrize("metric", get_metrics())
def test_metric_str(regres, metric, get_mats, get_targets):
    n_matrices, n_channels = 6, 3
    targets = get_targets(n_matrices)
    covmats = get_mats(n_matrices, n_channels, "spd")
    clf = regres(metric=metric)

    if regres is SVR and metric not in ['riemann', 'euclid', 'logeuclid']:
        with pytest.raises(ValueError):
            clf.fit(covmats, targets).predict(covmats)

    else:
        clf.fit(covmats, targets).predict(covmats)


@pytest.mark.parametrize("dist", ["not_real", 42])
def test_knn_dict_dist(dist, get_mats, get_targets):
    n_matrices, n_channels = 6, 3
    targets = get_targets(n_matrices)
    covmats = get_mats(n_matrices, n_channels, "spd")
    with pytest.raises(KeyError):
        clf = KNearestNeighborRegressor(metric={"distance": dist})
        clf.fit(covmats, targets).predict(covmats)


def test_1NN(get_mats, get_targets):
    """Test KNearestNeighborRegressor with K=1"""
    n_matrices, n_channels = 9, 3
    covmats = get_mats(n_matrices, n_channels, "spd")
    targets = get_targets(n_matrices)

    knn = KNearestNeighborRegressor(1, metric="riemann")
    knn.fit(covmats, targets)
    preds = knn.predict(covmats[:-1])
    assert_array_equal(targets[:-1], preds)


def test_svr_params():
    rsvr = SVR()
    assert rsvr.metric == 'riemann'

    rsvr.set_params(**{'metric': 'logeuclid'})
    assert rsvr.metric == 'logeuclid'

    rsvr.set_params(**{'max_iter': 501})
    assert rsvr.max_iter == 501


def test_svr_params_error(get_mats, get_targets):
    n_matrices, n_channels = 6, 3
    covmats = get_mats(n_matrices, n_channels, "spd")
    targets = get_targets(n_matrices)

    with pytest.raises(TypeError):
        SVR(C='hello').fit(covmats, targets)

    with pytest.raises(TypeError):
        SVR(foo=5)


@pytest.mark.parametrize("metric", ["riemann", "euclid", "logeuclid"])
def test_svr_cref_metric(get_mats, get_targets, metric):
    n_matrices, n_channels = 6, 3
    covmats = get_mats(n_matrices, n_channels, "spd")
    targets = get_targets(n_matrices)
    Cref = mean_covariance(covmats, metric=metric)

    rsvc = SVR(Cref=Cref).fit(covmats, targets)
    rsvc_1 = SVR(Cref=None, metric=metric).fit(covmats, targets)
    assert np.array_equal(rsvc.Cref_, rsvc_1.Cref_)


@pytest.mark.parametrize("metric", ["riemann", "euclid", "logeuclid"])
def test_svc_cref_callable(get_mats, get_targets, metric):
    n_matrices, n_channels = 6, 3
    covmats = get_mats(n_matrices, n_channels, "spd")
    targets = get_targets(n_matrices)
    def Cref(X): return mean_covariance(X, metric=metric)

    rsvc = SVR(Cref=Cref).fit(covmats, targets)
    rsvc_1 = SVR(metric=metric).fit(covmats, targets)
    assert np.array_equal(rsvc.Cref_, rsvc_1.Cref_)

    rsvc = SVR(Cref=Cref).fit(covmats, targets)
    rsvc.predict(covmats)
    rsvc_1 = SVR(metric=metric).fit(covmats, targets)
    rsvc_1.predict(covmats)
    assert np.array_equal(rsvc.Cref_, rsvc_1.Cref_)


@pytest.mark.parametrize("metric", ["riemann", "euclid", "logeuclid"])
def test_svc_cref_error(get_mats, get_targets, metric):
    n_matrices, n_channels = 6, 3
    covmats = get_mats(n_matrices, n_channels, "spd")
    targets = get_targets(n_matrices)
    def Cref(X, met): mean_covariance(X, metric=met)

    with pytest.raises(TypeError):
        SVR(Cref=Cref).fit(covmats, targets)

    Cref = metric

    with pytest.raises(TypeError):
        SVR(Cref=Cref).fit(covmats, targets)


@pytest.mark.parametrize("metric", ["riemann", "euclid", "logeuclid"])
def test_svc_kernel_callable(get_mats, get_targets, metric):
    n_matrices, n_channels = 6, 3
    covmats = get_mats(n_matrices, n_channels, "spd")
    targets = get_targets(n_matrices)

    rsvc = SVR(kernel_fct=kernel,
               metric=metric).fit(covmats, targets)
    rsvc_1 = SVR(metric=metric).fit(covmats, targets)
    p1 = rsvc.predict(covmats[:-1])
    p2 = rsvc_1.predict(covmats[:-1])
    assert np.array_equal(p1, p2)

    def custom_kernel(X, Y, Cref, metric):
        return np.ones((len(X), len(Y)))
    SVR(kernel_fct=custom_kernel,
        metric=metric).fit(covmats, targets).predict(covmats[:-1])

    def custom_kernel(X, Y, Cref):
        return np.ones((len(X), len(Y)))
    with pytest.raises(TypeError):
        SVR(kernel_fct=custom_kernel, metric=metric).fit(covmats, targets)

    custom_kernel = np.array([1, 2])
    with pytest.raises(TypeError):
        SVR(kernel_fct=custom_kernel, metric=metric).fit(covmats, targets)

    # check if pickleable
    pickle.dumps(rsvc)
    pickle.dumps(rsvc_1)



import numpy as np
from numpy.testing import assert_array_equal
import pytest

from conftest import get_metrics
from pyriemann.spatialfilters import Xdawn, CSP, SPoC, BilinearFilter, AJDC


@pytest.mark.parametrize("spfilt", [Xdawn, CSP, SPoC, BilinearFilter, AJDC])
@pytest.mark.parametrize("n_channels", [3, 5, 7])
@pytest.mark.parametrize("n_classes", [2, 3])
def test_spatial_filters(spfilt, n_channels,
                         get_mats, rndstate, get_labels, n_classes):
    if n_classes == 2:
        n_matrices, n_times = 10, 256
    else:
        n_matrices, n_times = 9, 256
    labels = get_labels(n_matrices, n_classes)
    if spfilt is Xdawn:
        X = rndstate.randn(n_matrices, n_channels, n_times)
    elif spfilt in (CSP, SPoC, BilinearFilter):
        X = get_mats(n_matrices, n_channels, "spd")
    elif spfilt is AJDC:
        n_subjects, n_conditions = 2, 2
        X = rndstate.randn(n_subjects, n_conditions, n_channels, n_times)

    clf_fit(spfilt, X, labels, n_channels, n_times)
    clf_fit_independence(spfilt, X, labels, n_channels)
    if spfilt is CSP:
        clf_fit_error(spfilt, X, labels)
    clf_transform(spfilt, X, labels, n_matrices, n_channels, n_times)
    if spfilt in (CSP, SPoC, BilinearFilter):
        clf_transform_error(spfilt, X, labels, n_channels)


def clf_fit(spfilt, X, labels, n_channels, n_times):
    n_classes = len(np.unique(labels))
    if spfilt is BilinearFilter:
        n_filters = 4
        sf = spfilt(filters=np.eye(n_filters, n_channels))
    elif spfilt is AJDC:
        sf = spfilt(dim_red={"n_components": n_channels - 1})
    else:
        sf = spfilt()

    sf.fit(X, labels)

    if spfilt is AJDC:
        assert sf.forward_filters_.shape == (sf.n_sources_, n_channels)
    elif spfilt is BilinearFilter:
        assert sf.filters_.shape == (n_filters, n_channels)
    elif spfilt is Xdawn:
        n_components = min(n_channels, sf.nfilter)
        assert len(sf.classes_) == n_classes
        assert sf.filters_.shape == (n_classes * n_components, n_channels)
        assert sf.patterns_.shape == (n_classes * n_components, n_channels)
        assert sf.evokeds_.shape == (n_classes * n_components, n_times)
    elif spfilt in [CSP, SPoC]:
        n_components = min(n_channels, sf.nfilter)
        assert sf.filters_.shape == (n_components, n_channels)
        assert sf.patterns_.shape == (n_components, n_channels)


def clf_fit_error(spfilt, X, labels):
    sf = spfilt()
    with pytest.raises(ValueError):
        sf.fit(X, labels * 0.0)  # 1 class
    with pytest.raises(ValueError):
        sf.fit(X, labels[:1])  # unequal # of samples
    with pytest.raises(TypeError):
        sf.fit(X, "foo")  # y must be an array
    with pytest.raises(TypeError):
        sf.fit("foo", labels)  # X must be an array
    with pytest.raises(ValueError):
        sf.fit(X[:, 0], labels)
    with pytest.raises(ValueError):
        sf.fit(X, X)


def clf_transform(spfilt, X, labels, n_matrices, n_channels, n_times):
    n_classes = len(np.unique(labels))
    if spfilt is BilinearFilter:
        n_filters = 4
        sf = spfilt(filters=np.eye(n_filters, n_channels))
    elif spfilt is AJDC:
        sf = spfilt(dim_red={"expl_var": 0.9})
    else:
        sf = spfilt()
    if spfilt is AJDC:
        sf.fit(X, labels)
        X_new = np.squeeze(X[0])
        n_matrices = X_new.shape[0]
        Xtr = sf.transform(X_new)
    else:
        Xtr = sf.fit(X, labels).transform(X)

    if spfilt is AJDC:
        assert Xtr.shape == (n_matrices, n_channels, n_times)
    elif spfilt is BilinearFilter:
        assert Xtr.shape == (n_matrices, n_filters, n_filters)
    elif spfilt is Xdawn:
        n_components = min(n_channels, sf.nfilter)
        assert Xtr.shape == (n_matrices, n_classes * n_components, n_times)
    else:
        n_components = min(n_channels, sf.nfilter)
        assert Xtr.shape == (n_matrices, n_components)


def clf_transform_error(spfilt, X, labels, n_channels):
    if spfilt is BilinearFilter:
        sf = spfilt(np.eye(n_channels))
    elif spfilt is AJDC:
        sf = spfilt(dim_red={"max_cond": 10})
    else:
        sf = spfilt()
    with pytest.raises(ValueError):
        sf.fit(X, labels).transform(X[:, :-1, :-1])


def clf_fit_independence(spfilt, X, labels, n_channels):
    if spfilt is BilinearFilter:
        sf = spfilt(np.eye(n_channels))
    elif spfilt is AJDC:
        sf = spfilt(dim_red={"max_cond": 10})
    else:
        sf = spfilt()
    sf.fit(X, labels)
    if spfilt is Xdawn:
        X_new = X[:, :-1, :]
    elif spfilt in (CSP, SPoC, BilinearFilter):
        X_new = X[:, :-1, :-1]
    elif spfilt is AJDC:
        X_new = X[:, :, :-1, :]
    # retraining with different size should erase previous fit
    sf.fit(X_new, labels)


@pytest.mark.parametrize("n_channels", [3, 4, 5])
@pytest.mark.parametrize("baseline_cov", [True, False])
def test_xdawn_baselinecov(n_channels, baseline_cov, rndstate, get_labels):
    n_classes, n_matrices, n_times = 2, 6, 100
    x = rndstate.randn(n_matrices, n_channels, n_times)
    labels = get_labels(n_matrices, n_classes)
    if baseline_cov:
        baseline_cov = np.identity(n_channels)
    else:
        baseline_cov = None
    xd = Xdawn(baseline_cov=baseline_cov)
    xd.fit(x, labels).transform(x)
    n_components = min(n_channels, xd.nfilter)
    assert xd.filters_.shape == (n_classes * n_components, n_channels)


@pytest.mark.parametrize("n_filters", [3, 4, 5])
@pytest.mark.parametrize("metric", get_metrics())
@pytest.mark.parametrize("log", [True, False])
def test_csp_init(n_filters, metric, log, get_mats, get_labels):
    n_classes, n_matrices, n_channels = 2, 6, 4
    covmats = get_mats(n_matrices, n_channels, "spd")
    labels = get_labels(n_matrices, n_classes)
    csp = CSP(nfilter=n_filters, metric=metric, log=log)
    Xtr = csp.fit(covmats, labels).transform(covmats)
    n_components = min(n_channels, n_filters)
    if log:
        assert Xtr.shape == (n_matrices, n_components)
    else:
        assert Xtr.shape == (n_matrices, n_components, n_components)
    assert csp.filters_.shape == (n_components, n_channels)
    assert csp.patterns_.shape == (n_components, n_channels)


def test_bilinearfilter_filter_error(get_mats, get_labels):
    n_classes, n_matrices, n_channels = 2, 6, 3
    covmats = get_mats(n_matrices, n_channels, "spd")
    labels = get_labels(n_matrices, n_classes)
    with pytest.raises(TypeError):
        BilinearFilter("foo").fit(covmats, labels)


def test_bilinearfilter_log_error(get_mats, get_labels):
    n_classes, n_matrices, n_channels = 2, 6, 3
    covmats = get_mats(n_matrices, n_channels, "spd")
    labels = get_labels(n_matrices, n_classes)
    with pytest.raises(TypeError):
        BilinearFilter(np.eye(3), log="foo").fit(covmats, labels)


@pytest.mark.parametrize("n_filters", [3, 4])
@pytest.mark.parametrize("log", [True, False])
def test_bilinearfilter_log(n_filters, log, get_mats, get_labels):
    n_classes, n_matrices, n_channels = 2, 6, 4
    covmats = get_mats(n_matrices, n_channels, "spd")
    labels = get_labels(n_matrices, n_classes)
    bf = BilinearFilter(np.eye(n_filters, n_channels), log=log)
    Xtr = bf.fit(covmats, labels).transform(covmats)
    if log:
        assert Xtr.shape == (n_matrices, n_filters)
    else:
        assert Xtr.shape == (n_matrices, n_filters, n_filters)


def test_ajdc_init():
    ajdc = AJDC(fmin=1, fmax=32, fs=64)
    assert ajdc.window == 128
    assert ajdc.overlap == 0.5
    assert ajdc.dim_red is None
    assert ajdc.verbose


def test_ajdc_fit(rndstate):
    n_subjects, n_conditions, n_channels, n_times = 5, 3, 8, 512
    X = rndstate.randn(n_subjects, n_conditions, n_channels, n_times)
    ajdc = AJDC(dim_red={"n_components": n_channels - 1}).fit(X)
    assert ajdc.forward_filters_.shape == (ajdc.n_sources_, n_channels)
    assert ajdc.backward_filters_.shape == (n_channels, ajdc.n_sources_)
    with pytest.warns(UserWarning):  # dim_red is None
        AJDC().fit(X)


def test_ajdc_fit_error(rndstate):
    n_subjects, n_conditions, n_channels, n_times = 2, 3, 8, 512
    ajdc = AJDC(dim_red={"expl_var": 0.9})
    with pytest.raises(ValueError):  # unequal # of conditions
        ajdc.fit(
            [
                rndstate.randn(n_conditions, n_channels, n_times),
                rndstate.randn(n_conditions + 1, n_channels, n_times),
            ]
        )
    with pytest.raises(ValueError):  # unequal # of channels
        ajdc.fit(
            [
                rndstate.randn(n_conditions, n_channels, n_times),
                rndstate.randn(n_conditions, n_channels + 1, n_times),
            ]
        )
    X = rndstate.randn(n_subjects, n_conditions, n_channels, n_times)
    V = rndstate.randn(n_channels, n_channels - 1)
    ajdc = AJDC(dim_red={'warm_restart': V})
    with pytest.raises(ValueError):  # initial diag not square
        ajdc.fit(X)


def test_ajdc_transform_error(rndstate):
    n_subjects, n_conditions, n_channels, n_times = 2, 2, 4, 256
    X = rndstate.randn(n_subjects, n_conditions, n_channels, n_times)
    ajdc = AJDC(dim_red={"warm_restart": np.eye(n_channels - 1)}).fit(X)
    n_matrices = 4
    X_new = rndstate.randn(n_matrices, n_channels, n_times)
    with pytest.raises(ValueError):  # not 3 dims
        ajdc.transform(X_new[0])
    with pytest.raises(ValueError):  # unequal # of chans
        ajdc.transform(rndstate.randn(n_matrices, n_channels + 1, 1))


def test_ajdc_fit_variable_input(rndstate):
    n_subjects, n_cond, n_chan, n_times = 2, 2, 3, 256
    X = rndstate.randn(n_subjects, n_cond, n_chan, n_times)
    ajdc = AJDC(dim_red={"expl_var": 0.9})
    # 3 subjects, same # conditions and channels, different # of times
    X = [
        rndstate.randn(n_cond, n_chan, n_times + rndstate.randint(500))
        for _ in range(3)
    ]
    ajdc.fit(X)

    # 2 subjects, 2 conditions, same # channels, different # of times
    X = [
        [rndstate.randn(n_chan, n_times + rndstate.randint(500))
         for _ in range(2)],
        [rndstate.randn(n_chan, n_times + rndstate.randint(500))
         for _ in range(2)],
    ]
    ajdc.fit(X)


def test_ajdc_inverse_transform(rndstate):
    n_subjects, n_conditions, n_channels, n_times = 2, 2, 5, 256
    X = rndstate.randn(n_subjects, n_conditions, n_channels, n_times)
    ajdc = AJDC(dim_red={"warm_restart": np.eye(n_channels - 1)}).fit(X)
    n_matrices = 4
    X_new = rndstate.randn(n_matrices, n_channels, n_times)
    Xt = ajdc.transform(X_new)
    Xit = ajdc.inverse_transform(Xt)
    assert_array_equal(Xit.shape, [n_matrices, n_channels, n_times])
    with pytest.raises(ValueError):  # not 3 dims
        ajdc.inverse_transform(Xt[0])
    with pytest.raises(ValueError):  # unequal # of sources
        shape = (n_matrices, ajdc.n_sources_ + 1, 1)
        ajdc.inverse_transform(rndstate.randn(*shape))

    Xit = ajdc.inverse_transform(Xt, supp=[ajdc.n_sources_ - 1])
    assert Xit.shape == (n_matrices, n_channels, n_times)
    with pytest.raises(ValueError):  # not a list
        ajdc.inverse_transform(Xt, supp=1)


def test_ajdc_get_src_expl_var(rndstate):
    n_subjects, n_conditions, n_channels, n_times = 2, 2, 3, 256
    X = rndstate.randn(n_subjects, n_conditions, n_channels, n_times)
    ajdc = AJDC(dim_red={"expl_var": 0.9}).fit(X)
    n_matrices = 4
    X_new = rndstate.randn(n_matrices, n_channels, n_times)
    v = ajdc.get_src_expl_var(X_new)
    assert v.shape == (n_matrices, ajdc.n_sources_)
    with pytest.raises(ValueError):  # not 3 dims
        ajdc.get_src_expl_var(X_new[0])
    with pytest.raises(ValueError):  # unequal # of chans
        ajdc.get_src_expl_var(rndstate.randn(n_matrices, n_channels + 1, 1))



import sys
import math
import numpy as np
from sklearn.model_selection import cross_val_score

from .utils.distance import distance, pairwise_distance
from .utils.mean import mean_covariance
from .utils.utils import check_metric
from .classification import MDM


def multiset_perm_number(y):
    """return the number of unique permutation in a multiset."""
    pr = 1
    for i in np.unique(y):
        pr *= math.factorial(np.sum(y == i))
    return math.factorial(len(y)) / pr


def unique_permutations(elements):
    """Return the list of unique permutations."""
    if len(elements) == 1:
        yield (elements[0], )
    else:
        unique_elements = set(elements)
        for first_element in unique_elements:
            remaining_elements = list(elements)
            remaining_elements.remove(first_element)
            for sub_permutation in unique_permutations(remaining_elements):
                yield (first_element, ) + sub_permutation


class BasePermutation():
    """Base object for permutations test"""

    def test(self, X, y, groups=None, verbose=True):
        """Performs the permutation test

        Parameters
        ----------
        X : array-like
            The data to fit. Can be, for example a list, or an array at
            least 2d.
        y : array-like
            The target variable to try to predict in the case of
            supervised learning.

        verbose : bool, default=True
            If true, print progress.
        """
        Npe = multiset_perm_number(y)
        self.scores_ = np.zeros(np.min([self.n_perms, int(Npe)]))

        # initial fit. This is usefull for transform data or for estimating
        # parameter that does not change across permutation, like the mean of
        # all data, the pairwise distance matrix, etc.
        X = self._initial_transform(X)

        # get the non permuted score
        self.scores_[0] = self.score(X, y, groups=groups)

        if Npe <= self.n_perms:
            print("Warning, number of unique permutations : %d" % Npe)
            perms = unique_permutations(y)
            ii = 0
            for perm in perms:
                if not np.array_equal(perm, y):
                    self.scores_[ii + 1] = self.score(X, perm, groups=groups)
                    ii += 1
                    if verbose:
                        self._print_progress(ii)

        else:
            rs = np.random.RandomState(self.random_state)
            for ii in range(self.n_perms - 1):
                perm = self._shuffle(y, groups, rs)
                self.scores_[ii + 1] = self.score(X, perm, groups=groups)
                if verbose:
                    self._print_progress(ii)
        if verbose:
            print("")
        self.p_value_ = (self.scores_[0] <= self.scores_).mean()

        return self.p_value_, self.scores_

    def _print_progress(self, ii):
        """Print permutation progress"""
        sys.stdout.write("Performing permutations : [%.1f%%]\r" %
                         ((100. * (ii + 1)) / self.n_perms))
        sys.stdout.flush()

    def _initial_transform(self, X):
        """Initial transformation. By default return X."""
        return X

    def _shuffle(self, y, groups, rs):
        """Return a shuffled copy of y eventually shuffle among same groups."""
        if groups is None:
            indices = rs.permutation(len(y))
        else:
            indices = np.arange(len(groups))
            for group in np.unique(groups):
                this_mask = (groups == group)
                indices[this_mask] = rs.permutation(indices[this_mask])
        return y[indices]

    def plot(self, nbins=10, range=None, axes=None):
        """Plot results of the permutation test.

        Parameters
        ----------
        nbins : integer or array_like or 'auto', default=10
            If an integer is given, bins + 1 bin edges are returned,
            consistently with np.histogram() for numpy version >= 1.3.
            Unequally spaced bins are supported if bins is a sequence.
        range : tuple or None, default=None
            The lower and upper range of the bins. Lower and upper outliers are
            ignored. If not provided, range is (x.min(), x.max()).
            Range has no effect if bins is a sequence.
            If bins is a sequence or range is specified, autoscaling is based
            on the specified bin range instead of the range of x.

        axes : axes handle, default=None
            Axes handle for matplotlib. if None a new figure will be created.
        """
        try:
            import matplotlib.pyplot as plt
        except ImportError:
            raise ImportError("Install matplotlib to plot permutation")

        if axes is None:
            fig, axes = plt.subplots(1, 1)
        axes.hist(self.scores_[1:], nbins, range)
        x_val = self.scores_[0]
        y_max = axes.get_ylim()[1]
        axes.plot([x_val, x_val], [0, y_max], '--r', lw=2)
        x_max = axes.get_xlim()[1]
        x_min = axes.get_xlim()[0]
        x_pos = x_min + ((x_max - x_min) * 0.25)
        axes.text(x_pos, y_max * 0.8, 'p-value: %.3f' % self.p_value_)
        axes.set_xlabel('Score')
        axes.set_ylabel('Count')
        return axes


class PermutationModel(BasePermutation):
    """Permutation test using any scikit-learn model for scoring.

    Perform a permutation test using the cross-validation score of any
    scikit-learn compatible model. Score is obtained with `cross_val_score`
    from scikit-learn. The score should be a "higer is better" metric.

    Parameters
    ----------
    n_perms : int, default=100
        The number of permutation. The minimum should be 20 for a resolution of
        0.05 p-value.
    model : sklearn compatible model, default=MDM()
        The model for scoring.
    cv : int or cross-validation generator or an iterable, default=3
        Determines the cross-validation splitting strategy.
        Possible inputs for cv are:

        - None, to use the default 3-fold cross validation;
        - integer, to specify the number of folds in a `(Stratified)KFold`;
        - an object to be used as a cross-validation generator;
        - an iterable yielding train, test splits.

        For integer/None inputs, if the estimator is a classifier and `y` is
        either binary or multiclass, `StratifiedKFold` is used. In all
        other cases, `KFold` is used.
    scoring : string or callable or None, default=None
        A string (see model evaluation documentation) or
        a scorer callable object / function with signature
        `scorer(estimator, X, y)`.
    n_jobs : integer, default=1
        The number of CPUs to use to do the computation. -1 means
        'all CPUs'.
    random_state : int, default=42
        random state for the permutation test.

    Attributes
    ----------
    p_value_ : float
        the p-value of the test
    scores_ : list
        contain all scores for all permutations. The fist element is the
        non-permuted score.

    See Also
    --------
    PermutationDistance
    """

    def __init__(self,
                 n_perms=100,
                 model=MDM(),
                 cv=3,
                 scoring=None,
                 n_jobs=1,
                 random_state=42):
        """Init."""
        self.n_perms = n_perms
        self.model = model
        self.cv = cv
        self.scoring = scoring
        self.n_jobs = n_jobs
        self.random_state = random_state

    def score(self, X, y, groups=None):
        """Score one permutation.

        Parameters
        ----------
        X : array-like
            The data to fit. Can be, for example a list, or an array at
            least 2d.

        y : array-like
            The target variable to try to predict in the case of
            supervised learning.
        """
        score = cross_val_score(
            self.model,
            X,
            y,
            cv=self.cv,
            n_jobs=self.n_jobs,
            scoring=self.scoring,
            groups=groups)
        return score.mean()


class PermutationDistance(BasePermutation):
    """Permutation test based on distance.

    Perform a permutation test based on distance. You have the choice of 3
    different statistic:

    - 'pairwise' :
        the statistic is based on paiwire distance as
        descibed in [1]_. This is the fastest option for low sample size since
        the pairwise distance matrix does not need to be estimated for each
        permutation.

    - 'ttest' :
        t-test based statistic obtained by the ration of the
        distance between each riemannian centroid and the group dispersion.
        The means have to be estimated for each permutation, leading to a
        slower procedure. However, this can be used for high sample size.

    - 'ftest':
        f-test based statistic estimated using the between and
        within group variability. As for the 'ttest' stats, group centroid
        are estimated for each permutation.

    Parameters
    ----------
    n_perms : int, default=100
        The number of permutation. The minimum should be 20 for a resolution of
        0.05 p-value.
    metric : string | dict, default="riemann"
        Metric used for mean estimation (for the list of supported metrics,
        see :func:`pyriemann.utils.mean.mean_covariance`) and
        for distance estimation
        (see :func:`pyriemann.utils.distance.distance`).
        The metric can be a dict with two keys, "mean" and "distance"
        in order to pass different metrics.
    mode : string, default='pairwise'
        Type of statistic to use. could be 'pairwise', 'ttest' of 'ftest'
    n_jobs : integer, default=1
        The number of CPUs to use to do the computation. -1 means
        'all CPUs'.
    random_state : int, default=42
        random state for the permutation test.
    estimator : None or sklearn compatible estimator, default=None
        If provided, data are transformed before every permutation. should
        not be used unless a supervised opperation must be applied on the data.
        This would be the case for ERP covariance.

    Attributes
    ----------
    p_value_ : float
        the p-value of the test
    scores_ : list
        contain all scores for all permutations. The fist element is the
        non-permuted score.

    See Also
    --------
    PermutationModel

    References
    ----------
    .. [1] `A new method for non-parametric multivariate analysis of variance
        <https://doi.org/10.1111/j.1442-9993.2001.01070.pp.x>`_
        M. Anderson. Austral ecology, Volume 26, Issue 1, February 2001.
    """

    def __init__(self,
                 n_perms=100,
                 metric="riemann",
                 mode='pairwise',
                 n_jobs=1,
                 random_state=42,
                 estimator=None):
        """Init."""
        self.n_perms = n_perms
        if mode not in ['pairwise', 'ttest', 'ftest']:
            raise (ValueError("mode must be 'pairwise', 'ttest' or 'ftest'"))
        self.mode = mode
        self.metric = metric
        self.n_jobs = n_jobs
        self.random_state = random_state
        self.estimator = estimator

    def score(self, X, y, groups=None):
        """Score of a permutation.

        Parameters
        ----------
        X : array-like
            The data to fit. Can be, for example a list, or an array at
            least 2d.
        y : array-like
            The target variable to try to predict in the case of
            supervised learning.
        """
        if self.estimator:
            X = self.estimator.fit_transform(X, y)
            X = self.__init_transform(X)

        if self.mode == 'ttest':
            return self._score_ttest(X, y)
        elif self.mode == 'ftest':
            return self._score_ftest(X, y)
        elif self.mode == 'pairwise':
            return self._score_pairwise(X, y)

    def _initial_transform(self, X):
        """Initial transform"""
        # if an estimator provided, then transform w
        if self.estimator:
            return X

        return self.__init_transform(X)

    def __init_transform(self, X):
        """Init tr"""
        self.mdm = MDM(metric=self.metric, n_jobs=self.n_jobs)
        self.mdm.metric_mean, self.mdm.metric_dist = check_metric(self.metric)
        if self.mode == 'ftest':
            self.global_mean = mean_covariance(X, metric=self.mdm.metric_mean)
        elif self.mode == 'pairwise':
            X = pairwise_distance(X, metric=self.mdm.metric_dist, squared=True)
        return X

    def _score_ftest(self, X, y):
        """Get the score"""
        mdm = self.mdm.fit(X, y)
        covmeans = np.array(mdm.covmeans_)

        # estimates between classes variability
        n_classes = len(covmeans)
        between = 0
        for ix, classe in enumerate(mdm.classes_):
            di = distance(
                covmeans[ix],
                self.global_mean,
                metric=mdm.metric_dist,
                squared=True,
            )
            between += np.sum(y == classe) * di
        between /= (n_classes - 1)

        # estimates within class variability
        within = 0
        for ix, classe in enumerate(mdm.classes_):
            within += distance(
                X[y == classe],
                covmeans[ix],
                metric=mdm.metric_dist,
                squared=True,
            ).sum()
        within /= (len(y) - n_classes)

        score = between / within
        return score

    def _score_ttest(self, X, y):
        """Get the score"""
        mdm = self.mdm.fit(X, y)
        covmeans = np.array(mdm.covmeans_)

        # estimates distances between means
        n_classes = len(covmeans)
        pairs = pairwise_distance(covmeans, metric=mdm.metric_dist)
        mean_dist = np.triu(pairs).sum()
        mean_dist /= (n_classes * (n_classes - 1)) / 2.0

        dist = 0
        for ix, classe in enumerate(mdm.classes_):
            di = distance(
                X[y == classe],
                covmeans[ix],
                metric=mdm.metric_dist,
                squared=True,
            ).mean()
            dist += (di / np.sum(y == classe))
        score = mean_dist / np.sqrt(dist)
        return score

    def _score_pairwise(self, X, y):
        """Score for the pairwise distance test."""
        classes = np.unique(y)
        n_classes = len(classes)
        n_samples = len(y)
        total_ss = X.sum() / (2 * n_samples)
        pattern = np.zeros((n_samples, n_samples))
        for classe in classes:
            ix = (y == classe)
            pattern += (np.outer(ix, ix) / ix.sum())

        within_ss = (X * pattern).sum() / 2

        between_ss = total_ss - within_ss

        score = ((between_ss / (n_classes - 1)) / (within_ss /
                                                   (n_samples - n_classes)))

        return score



import numpy as np
from numpy.testing import assert_array_equal, assert_array_almost_equal
import pytest

from pyriemann.utils.base import logm
from pyriemann.utils.kernel import (
    kernel,
    kernel_euclid,
    kernel_logeuclid,
    kernel_riemann
)
from pyriemann.utils.mean import mean_covariance
from pyriemann.utils.test import is_sym_pos_semi_def as is_spsd

rker_str = ['euclid', 'logeuclid', 'riemann']
rker_fct = [kernel_euclid, kernel_logeuclid, kernel_riemann]


@pytest.mark.parametrize("ker", rker_fct)
def test_kernel_x_x(ker, get_mats):
    """Test kernel build"""
    n_matrices, n_channels = 7, 3
    X = get_mats(n_matrices, n_channels, "spd")
    K = ker(X, X)
    assert K.shape == (n_matrices, n_matrices)
    assert is_spsd(K)
    assert_array_almost_equal(K, ker(X))


@pytest.mark.parametrize("ker", rker_str)
def test_kernel_cref(ker, get_mats):
    """Test kernel reference"""
    n_matrices, n_channels = 5, 3
    X = get_mats(n_matrices, n_channels, "spd")
    cref = mean_covariance(X, metric=ker)
    K = kernel(X, X, metric=ker)
    K1 = kernel(X, X, Cref=cref, metric=ker)
    assert_array_equal(K, K1)


@pytest.mark.parametrize("ker", rker_str)
def test_kernel_x_y(ker, get_mats):
    """Test kernel for different X and Y"""
    n_matrices_X, n_matrices_Y, n_channels = 6, 5, 3
    X = get_mats(n_matrices_X, n_channels, "spd")
    Y = get_mats(n_matrices_Y, n_channels, "spd")
    K = kernel(X, Y, metric=ker)
    assert K.shape == (n_matrices_X, n_matrices_Y)


@pytest.mark.parametrize("ker", rker_str)
def test_metric_string(ker, get_mats):
    """Test generic kernel function"""
    n_matrices, n_channels = 5, 3
    X = get_mats(n_matrices, n_channels, "spd")
    K = globals()[f'kernel_{ker}'](X)
    K1 = kernel(X, metric=ker)
    assert_array_equal(K, K1)


def test_metric_string_error(get_mats):
    """Test generic kernel function error raise"""
    n_matrices, n_channels = 5, 3
    X = get_mats(n_matrices, n_channels, "spd")
    with pytest.raises(ValueError):
        kernel(X, metric='foo')


@pytest.mark.parametrize("ker", rker_str)
def test_input_dimension_error(ker, get_mats):
    """Test errors for incorrect dimension"""
    n_matrices, n_channels = 5, 3
    X = get_mats(n_matrices, n_channels, "spd")
    Y = get_mats(n_matrices, n_channels + 1, "spd")
    cref = get_mats(1, n_channels + 1, "spd")[0]
    if ker == 'riemann':
        with pytest.raises(AssertionError):
            kernel(X, Cref=cref, metric=ker)
    with pytest.raises(AssertionError):
        kernel(X, Y, metric=ker)


@pytest.mark.parametrize("n_dim0, n_dim1", [(4, 4), (4, 5), (5, 4)])
def test_euclid(n_dim0, n_dim1, rndstate):
    """Test Euclidean kernel for generic matrices"""
    n_matrices_X, n_matrices_Y = 2, 3
    X = rndstate.randn(n_matrices_X, n_dim0, n_dim1)
    Y = rndstate.randn(n_matrices_Y, n_dim0, n_dim1)
    K = kernel_euclid(X, Y)
    assert K.shape == (n_matrices_X, n_matrices_Y)

    K1 = np.empty((n_matrices_X, n_matrices_Y))
    K2 = np.empty((n_matrices_X, n_matrices_Y))
    for i in range(n_matrices_X):
        for j in range(n_matrices_Y):
            K1[i, j] = np.trace(X[i].T @ Y[j])
            K2[i, j] = np.dot(X[i].flatten(), Y[j].flatten())
    assert_array_almost_equal(K, K1)
    assert_array_almost_equal(K, K2)


def test_riemann_correctness(get_mats):
    """Test Riemannian kernel correctness"""
    n_matrices, n_channels = 5, 3
    X = get_mats(n_matrices, n_channels, "spd")
    K = kernel_riemann(X, Cref=np.eye(n_channels), reg=0)

    log_X = logm(X)
    tensor = np.tensordot(log_X, log_X.T, axes=1)
    K1 = np.trace(tensor, axis1=1, axis2=2)
    assert_array_almost_equal(K, K1)



import numpy as np
import pytest
from pytest import approx
from scipy.stats import gmean

from pyriemann.utils.geodesic import geodesic_riemann
from pyriemann.utils.mean import (
    mean_covariance,
    mean_ale,
    mean_alm,
    mean_euclid,
    mean_harmonic,
    mean_identity,
    mean_kullback_sym,
    mean_logdet,
    mean_logeuclid,
    mean_power,
    mean_riemann,
    mean_wasserstein,
    maskedmean_riemann,
    nanmean_riemann,
)


@pytest.mark.parametrize("kind", ["spd", "hpd"])
@pytest.mark.parametrize(
    "mean",
    [
        mean_ale,
        mean_euclid,
        mean_harmonic,
        mean_identity,
        mean_kullback_sym,
        mean_logdet,
        mean_logeuclid,
        mean_power,
        mean_riemann,
        mean_wasserstein,
        nanmean_riemann,
    ],
)
def test_mean_shape(kind, mean, get_mats):
    """Test the shape of mean"""
    n_matrices, n_channels = 5, 3
    mats = get_mats(n_matrices, n_channels, kind)
    if mean is mean_ale and kind == "hpd":
        pytest.skip()
    if mean == mean_power:
        C = mean(mats, 0.42)
    else:
        C = mean(mats)
    assert C.shape == (n_channels, n_channels)


@pytest.mark.parametrize("kind", ["spd", "hpd"])
@pytest.mark.parametrize(
    "mean", [mean_logdet, mean_riemann, mean_wasserstein, nanmean_riemann]
)
def test_mean_shape_with_init(kind, mean, get_mats):
    """Test the shape of mean with init"""
    n_matrices, n_channels = 5, 3
    mats = get_mats(n_matrices, n_channels, kind)
    C = mean(mats, init=mats[0])
    assert C.shape == (n_channels, n_channels)


@pytest.mark.parametrize("kind", ["spd", "hpd"])
@pytest.mark.parametrize(
    "mean",
    [
        mean_euclid,
        mean_harmonic,
        mean_kullback_sym,
        mean_logdet,
        mean_logeuclid,
        mean_riemann,
        mean_wasserstein,
        nanmean_riemann,
    ],
)
def test_mean_weight_zero(kind, mean, get_mats):
    """Setting one weight to almost 0 it's almost like not passing the mat"""
    n_matrices, n_channels = 5, 3
    mats = get_mats(n_matrices, n_channels, kind)
    w = 2.3 * np.ones(n_matrices)
    C = mean(mats[1:], sample_weight=w[1:])
    w[0] = 1e-12
    Cw = mean(mats, sample_weight=w)
    assert C == approx(Cw, rel=1e-6, abs=1e-8)


@pytest.mark.parametrize(
    "mean",
    [
        mean_euclid,
        mean_harmonic,
        mean_kullback_sym,
        mean_logdet,
        mean_logeuclid,
        mean_riemann,
        mean_wasserstein,
        nanmean_riemann,
    ],
)
def test_mean_weight_len_error(mean, get_mats):
    n_matrices, n_channels = 3, 2
    mats = get_mats(n_matrices, n_channels, "spd")
    with pytest.raises(ValueError):
        mean(mats, sample_weight=np.ones(n_matrices + 1))


@pytest.mark.parametrize(
    "mean", [
        mean_ale,
        mean_alm,
        mean_logdet,
        mean_power,
        mean_riemann,
        mean_wasserstein,
        nanmean_riemann
    ]
)
def test_mean_warning_convergence(mean, get_mats):
    """Test warning for convergence not reached """
    n_matrices, n_channels = 3, 2
    mats = get_mats(n_matrices, n_channels, "spd")
    with pytest.warns(UserWarning):
        if mean == mean_power:
            mean(mats, 0.3, maxiter=0)
        else:
            mean(mats, maxiter=0)


@pytest.mark.parametrize("kind", ["spd", "hpd"])
@pytest.mark.parametrize(
    "mean",
    [
        mean_ale,
        mean_euclid,
        mean_harmonic,
        mean_identity,
        mean_kullback_sym,
        mean_logdet,
        mean_logeuclid,
        mean_power,
        mean_riemann,
        mean_wasserstein,
    ],
)
def test_mean_of_means(kind, mean, get_mats):
    """Test mean of submeans equal to grand mean"""
    n_matrices, n_channels = 10, 3
    mats = get_mats(n_matrices, n_channels, kind)
    if mean is mean_ale and kind == "hpd":
        pytest.skip()
    if mean == mean_power:
        p = -0.42
        C = mean(mats, p)
        C1 = mean(mats[:n_matrices//2], p)
        C2 = mean(mats[n_matrices//2:], p)
        C3 = mean(np.array([C1, C2]), p)
    else:
        C = mean(mats)
        C1 = mean(mats[:n_matrices//2])
        C2 = mean(mats[n_matrices//2:])
        C3 = mean(np.array([C1, C2]))
    assert C3 == approx(C, 6)


@pytest.mark.parametrize("kind", ["spd", "hpd"])
def test_mean_alm(kind, get_mats):
    """Test the ALM mean"""
    n_matrices, n_channels = 3, 3
    mats = get_mats(n_matrices, n_channels, kind)
    C_alm = mean_alm(mats)
    assert C_alm.shape == (n_channels, n_channels)
    C_riem = mean_riemann(mats)
    assert C_alm == approx(C_riem, abs=1e-6, rel=1e-3)


@pytest.mark.parametrize("kind", ["spd", "hpd"])
def test_mean_alm_2matrices(kind, get_mats):
    """Test the ALM mean with 2 matrices"""
    n_matrices, n_channels = 2, 3
    mats = get_mats(n_matrices, n_channels, kind)
    C = mean_alm(mats)
    assert np.all(C == geodesic_riemann(mats[0], mats[1], alpha=0.5))


@pytest.mark.parametrize("complex_valued", [True, False])
def test_mean_euclid(rndstate, complex_valued):
    """Test the Euclidean mean for generic matrices"""
    n_matrices, n_dim0, n_dim1 = 10, 3, 4
    mats = rndstate.randn(n_matrices, n_dim0, n_dim1)
    if complex_valued:
        mats = mats + 1j * rndstate.randn(n_matrices, n_dim0, n_dim1)
    assert mean_euclid(mats) == approx(mats.mean(axis=0))


def test_mean_identity(get_mats):
    """Test the identity mean"""
    n_matrices, n_channels = 2, 3
    mats = get_mats(n_matrices, n_channels, "spd")
    C = mean_identity(mats)
    assert np.all(C == np.eye(n_channels))


@pytest.mark.parametrize("kind", ["spd", "hpd"])
def test_mean_power(kind, get_mats):
    """Test the power mean"""
    n_matrices, n_channels = 3, 3
    mats = get_mats(n_matrices, n_channels, kind)
    assert mean_power(mats, 1) == approx(mean_euclid(mats))
    assert mean_power(mats, 0) == approx(mean_riemann(mats))
    assert mean_power(mats, -1) == approx(mean_harmonic(mats))


def test_mean_power_errors(get_mats):
    """Test the power mean errors"""
    n_matrices, n_channels = 3, 2
    mats = get_mats(n_matrices, n_channels, "spd")

    with pytest.raises(ValueError):  # exponent is not a scalar
        mean_power(mats, [1])
    with pytest.raises(ValueError):  # exponent is not in [-1,1]
        mean_power(mats, 3)


@pytest.mark.parametrize("kind", ["spd", "hpd"])
@pytest.mark.parametrize("init", [True, False])
def test_mean_riemann(kind, init, get_mats_params):
    """Test the Riemannian mean with same eigen vectors"""
    n_matrices, n_channels = 10, 3
    mats, eigvals, eigvecs = get_mats_params(n_matrices, n_channels, kind)
    if init:
        C = mean_riemann(mats, init=mats[0])
    else:
        C = mean_riemann(mats)
    eigval = np.exp(np.mean(np.log(eigvals), axis=0))
    Ctrue = eigvecs @ np.diag(eigval) @ eigvecs.conj().T
    assert C == approx(Ctrue)


@pytest.mark.parametrize("kind", ["spd", "hpd"])
def test_mean_riemann_properties(kind, get_mats):
    n_matrices, n_channels = 5, 3
    mats = get_mats(n_matrices, n_channels, kind)
    C = mean_riemann(mats)

    # congruence-invariance, P2 in [Moakher2005] or P6 in [Nakamura2009]
    W = np.random.normal(size=(n_channels, n_channels))  # must be invertible
    assert W @ C @ W.T == approx(mean_riemann(W @ mats @ W.T))

    # self-duality, P3 in [Moakher2005] or P8 in [Nakamura2009]
    assert C == approx(np.linalg.inv(mean_riemann(np.linalg.inv(mats))))

    # determinant identity, P9 in [Nakamura2009]
    assert np.linalg.det(C) == approx(gmean(np.linalg.det(mats)))


@pytest.mark.parametrize("init", [True, False])
def test_mean_masked_riemann_shape(init, get_mats, get_masks):
    """Test the masked Riemannian mean"""
    n_matrices, n_channels = 5, 3
    mats = get_mats(n_matrices, n_channels, "spd")
    masks = get_masks(n_matrices, n_channels)
    if init:
        C = maskedmean_riemann(mats, masks, tol=10e-3, init=mats[0])
    else:
        C = maskedmean_riemann(mats, masks, tol=10e-3)
    assert C.shape == (n_channels, n_channels)


@pytest.mark.parametrize("init", [True, False])
def test_mean_nan_riemann_shape(init, get_mats, rndstate):
    """Test the Riemannian NaN-mean"""
    n_matrices, n_channels = 10, 6
    mats = get_mats(n_matrices, n_channels, "spd")
    emean = np.mean(mats, axis=0)
    for i in range(n_matrices):
        corrup_channels = rndstate.choice(
            np.arange(0, n_channels), size=n_channels // 3, replace=False)
        for j in corrup_channels:
            mats[i, j] = np.nan
            mats[i, :, j] = np.nan
    if init:
        C = nanmean_riemann(mats, tol=10e-3, init=emean)
    else:
        C = nanmean_riemann(mats, tol=10e-3)
    assert C.shape == (n_channels, n_channels)


def test_mean_nan_riemann_errors(get_mats):
    """Test the Riemannian NaN-mean errors"""
    n_matrices, n_channels = 5, 4
    mats = get_mats(n_matrices, n_channels, "spd")

    with pytest.raises(ValueError):  # not symmetric NaN values
        mats_ = mats.copy()
        mats_[0, 0] = np.nan  # corrup only a row, not its corresp column
        nanmean_riemann(mats_)
    with pytest.raises(ValueError):  # not rows and columns NaN values
        mats_ = mats.copy()
        mats_[1, 0, 1] = np.nan  # corrup an off-diagonal value
        nanmean_riemann(mats_)


def callable_np_average(X, sample_weight=None):
    return np.average(X, axis=0, weights=sample_weight)


@pytest.mark.parametrize(
    "metric, mean",
    [
        ("ale", mean_ale),
        ("alm", mean_alm),
        ("euclid", mean_euclid),
        ("harmonic", mean_harmonic),
        ("identity", mean_identity),
        ("kullback_sym", mean_kullback_sym),
        ("logdet", mean_logdet),
        ("logeuclid", mean_logeuclid),
        ("riemann", mean_riemann),
        ("wasserstein", mean_wasserstein),
        (callable_np_average, mean_euclid),
    ],
)
def test_mean_covariance_metric(metric, mean, get_mats):
    """Test mean_covariance for metric"""
    n_matrices, n_channels = 3, 3
    mats = get_mats(n_matrices, n_channels, "spd")
    C = mean_covariance(mats, metric=metric)
    Ctrue = mean(mats)
    assert np.all(C == Ctrue)


def test_mean_covariance_args(get_mats):
    """Test mean_covariance with different arguments"""
    n_matrices, n_channels = 3, 3
    mats = get_mats(n_matrices, n_channels, "spd")
    mean_covariance(mats, metric='ale', maxiter=5)
    mean_covariance(mats, metric='logdet', tol=10e-3)
    mean_covariance(mats, metric='riemann', init=np.eye(n_channels))



import numpy as np
import pytest
from pytest import approx

from pyriemann.utils.geodesic import (
    geodesic,
    geodesic_euclid,
    geodesic_logeuclid,
    geodesic_riemann,
)
from pyriemann.utils.mean import (
    mean_euclid,
    mean_logeuclid,
    mean_riemann,
)


def get_geod_func():
    geod_func = [geodesic_euclid, geodesic_logeuclid, geodesic_riemann]
    for gf in geod_func:
        yield gf


def get_geod_name():
    geod_name = ["euclid", "logeuclid", "riemann"]
    for gn in geod_name:
        yield gn


def geodesic_0(gfun, A, B):
    assert gfun(A, B, 0) == approx(A)


def geodesic_1(gfun, A, B):
    assert gfun(A, B, 1) == approx(B)


def geodesic_middle(gfun, A, B, Ctrue):
    assert gfun(A, B, 0.5) == approx(Ctrue)


@pytest.mark.parametrize("gfun", get_geod_func())
def test_geodesic_all_simple(gfun):
    n_channels = 3
    if gfun is geodesic_euclid:
        A = 1.0 * np.eye(n_channels)
        B = 2.0 * np.eye(n_channels)
        Ctrue = 1.5 * np.eye(n_channels)
    else:
        A = 0.5 * np.eye(n_channels)
        B = 2 * np.eye(n_channels)
        Ctrue = np.eye(n_channels)
    geodesic_0(gfun, A, B)
    geodesic_1(gfun, A, B)
    geodesic_middle(gfun, A, B, Ctrue)


@pytest.mark.parametrize("kind", ["spd", "hpd"])
@pytest.mark.parametrize("gfun", get_geod_func())
def test_geodesic_all_random(kind, gfun, get_mats):
    n_matrices, n_channels = 2, 5
    mats = get_mats(n_matrices, n_channels, kind)
    A, B = mats[0], mats[1]
    if gfun is geodesic_euclid:
        Ctrue = mean_euclid(mats)
    elif gfun is geodesic_logeuclid:
        Ctrue = mean_logeuclid(mats)
    elif gfun is geodesic_riemann:
        Ctrue = mean_riemann(mats)
    geodesic_0(gfun, A, B)
    geodesic_1(gfun, A, B)
    geodesic_middle(gfun, A, B, Ctrue)


@pytest.mark.parametrize("complex_valued", [True, False])
def test_geodesic_euclid(rndstate, complex_valued):
    """Test Euclidean geodesic for generic matrices"""
    n_matrices, n_dim0, n_dim1 = 2, 3, 4
    mats = rndstate.randn(n_matrices, n_dim0, n_dim1)
    if complex_valued:
        mats = mats + 1j * rndstate.randn(n_matrices, n_dim0, n_dim1)
    A, B = mats[0], mats[1]
    assert geodesic_euclid(A, B).shape == A.shape


@pytest.mark.parametrize("metric", get_geod_name())
def test_geodesic_wrapper_ndarray(metric, get_mats):
    n_matrices, n_channels = 5, 3
    A = get_mats(n_matrices, n_channels, "spd")
    B = get_mats(n_matrices, n_channels, "spd")
    assert geodesic(A[0], B[0], .3, metric=metric).shape == A[0].shape
    assert geodesic(A, B, .2, metric=metric).shape == A.shape  # 3D arrays

    n_sets = 4
    C = np.asarray([A for _ in range(n_sets)])
    D = np.asarray([B for _ in range(n_sets)])
    assert geodesic(C, D, .7, metric=metric).shape == C.shape  # 4D arrays


@pytest.mark.parametrize("metric", get_geod_name())
def test_geodesic_wrapper_simple(metric):
    n_channels = 3
    A = 0.5 * np.eye(n_channels)
    if metric == "euclid":
        B = 1.5 * np.eye(n_channels)
    else:
        B = 2.0 * np.eye(n_channels)
    assert geodesic(A, B, 0.5, metric=metric) == approx(np.eye(n_channels))


@pytest.mark.parametrize("metric, gfun", zip(get_geod_name(), get_geod_func()))
def test_geodesic_wrapper_random(metric, gfun, get_mats):
    n_matrices, n_channels = 2, 5
    mats = get_mats(n_matrices, n_channels, "spd")
    A, B = mats[0], mats[1]
    if gfun is geodesic_euclid:
        Ctrue = mean_euclid(mats)
    elif gfun is geodesic_logeuclid:
        Ctrue = mean_logeuclid(mats)
    elif gfun is geodesic_riemann:
        Ctrue = mean_riemann(mats)
    assert geodesic(A, B, 0.5, metric=metric) == approx(Ctrue)



"""Spatial filtering function."""
import warnings

import numpy as np
from scipy.linalg import eigh, inv
from sklearn.base import BaseEstimator, TransformerMixin

from .utils.covariance import normalize, get_nondiag_weight, cov_est_functions
from .utils.mean import mean_covariance
from .utils.utils import check_function
from .utils.ajd import ajd_pham
from . import estimation as est
from .preprocessing import Whitening


class Xdawn(BaseEstimator, TransformerMixin):
    """Xdawn algorithm.

    Xdawn [1]_ is a spatial filtering method designed to improve the signal
    to signal + noise ratio (SSNR) of the ERP responses. Xdawn was originaly
    designed for P300 evoked potential by enhancing the target response with
    respect to the non-target response [2]_. This implementation is a
    generalization to any type of ERP.

    Parameters
    ----------
    nfilter : int, default=4
        The number of components to decompose M/EEG signals.
    classes : list of int | None, default=None
        List of classes to take into account for Xdawn.
        If None, all classes will be accounted.
    estimator : string, default='scm'
        Covariance matrix estimator, see
        :func:`pyriemann.utils.covariance.covariances`.
    baseline_cov : None | array, shape(n_channels, n_channels), default=None
        Covariance matrix to which the average signals are compared. If None,
        the baseline covariance is computed across all trials and time samples.

    Attributes
    ----------
    classes_ : ndarray, shape (n_classes,)
        Labels for each class.
    filters_ : ndarray, shape (n_classes x min(n_channels, n_filters), \
            n_channels)
        If fit, the Xdawn components used to decompose the data for each event
        type, concatenated.
    patterns_ : ndarray, shape (n_classes x min(n_channels, n_filters), \
            n_channels)
        If fit, the Xdawn patterns used to restore M/EEG signals for each event
        type, concatenated.
    evokeds_ : ndarray, shape (n_classes x min(n_channels, n_filters), n_times)
        If fit, the evoked response for each event type, concatenated.

    See Also
    --------
    XdawnCovariances

    References
    ----------
    .. [1] `xDAWN algorithm to enhance evoked potentials: application to
        brain-computer interface
        <https://hal.archives-ouvertes.fr/hal-00454568/fr/>`_
        B. Rivet, A. Souloumiac, V. Attina, and G. Gibert. IEEE Transactions on
        Biomedical Engineering, 2009, 56 (8), pp.2035-43.
    .. [2] `Theoretical analysis of xDAWN algorithm: application to an
        efficient sensor selection in a P300 BCI
        <https://hal.archives-ouvertes.fr/hal-00619997>`_
        B. Rivet, H. Cecotti, A. Souloumiac, E. Maby, J. Mattout. EUSIPCO 2011
        19th European Signal Processing Conference, Aug 2011, Barcelone, Spain.
        pp.1382-1386.
    """

    def __init__(self, nfilter=4, classes=None, estimator='scm',
                 baseline_cov=None):
        """Init."""
        self.nfilter = nfilter
        self.classes = classes
        self.estimator = estimator
        self.baseline_cov = baseline_cov

    @property
    def estimator_fn(self):
        return check_function(self.estimator, cov_est_functions)

    def fit(self, X, y):
        """Train Xdawn spatial filters.

        Parameters
        ----------
        X : ndarray, shape (n_trials, n_channels, n_times)
            Set of trials.
        y : ndarray, shape (n_trials,)
            Labels for each trial.

        Returns
        -------
        self : Xdawn instance
            The Xdawn instance.
        """
        n_trials, n_channels, n_times = X.shape

        self.classes_ = (np.unique(y) if self.classes is None else
                         self.classes)

        Cx = self.baseline_cov
        if Cx is None:
            tmp = X.transpose((1, 2, 0))
            Cx = np.asarray(self.estimator_fn(
                tmp.reshape(n_channels, n_times * n_trials)
            ))

        self.evokeds_ = []
        self.filters_ = []
        self.patterns_ = []
        for c in self.classes_:
            # Prototyped response for each class
            P = np.mean(X[y == c], axis=0)

            # Covariance matrix of the prototyper response & signal
            C = np.asarray(self.estimator_fn(P))

            # Spatial filters
            evals, evecs = eigh(C, Cx)
            evecs = evecs[:, np.argsort(evals)[::-1]]  # sort eigenvectors
            evecs /= np.apply_along_axis(np.linalg.norm, 0, evecs)
            V = evecs
            A = np.linalg.pinv(V.T)
            # create the reduced prototyped response
            self.filters_.append(V[:, 0:self.nfilter].T)
            self.patterns_.append(A[:, 0:self.nfilter].T)
            self.evokeds_.append(V[:, 0:self.nfilter].T @ P)

        self.evokeds_ = np.concatenate(self.evokeds_, axis=0)
        self.filters_ = np.concatenate(self.filters_, axis=0)
        self.patterns_ = np.concatenate(self.patterns_, axis=0)
        return self

    def transform(self, X):
        """Apply spatial filters.

        Parameters
        ----------
        X : ndarray, shape (n_trials, n_channels, n_times)
            Set of trials.

        Returns
        -------
        Xf : ndarray, shape (n_trials, n_classes x min(n_channels, n_filters),\
                n_times)
            Set of spatialy filtered trials.
        """
        X = self.filters_ @ X
        return X


class BilinearFilter(BaseEstimator, TransformerMixin):
    r"""Bilinear spatial filter.

    Bilinear spatial filter for SPD matrices allows to define a custom spatial
    filter for bilinear projection of the data:

    .. math::
        \mathbf{Xf}_i = \mathbf{V} \mathbf{X}_i \mathbf{V}^T

    If log parameter is set to true, will return the log of the diagonal:

    .. math::
        \mathbf{xf}_i = \log ( \mathrm{diag} (\mathbf{Xf}_i) )

    Parameters
    ----------
    filters : ndarray, shape (n_filters, n_channels)
        The filters for bilinear transform.
    log : bool, default=False
        If true, return the log variance, otherwise return the spatially
        filtered covariance matrices.

    Attributes
    ----------
    filters_ : ndarray, shape (n_filters, n_channels)
        If fit, the filter components used to decompose the data for each event
        type, concatenated.
    """

    def __init__(self, filters, log=False):
        """Init."""
        self.filters_ = filters
        self.filters = filters
        self.log = log

    def fit(self, X, y):
        """Train BilinearFilter spatial filters.

        Parameters
        ----------
        X : ndarray, shape (n_trials, n_channels, n_channels)
            Set of covariance matrices.
        y : ndarray, shape (n_trials,)
            Labels for each trial.

        Returns
        -------
        self : BilinearFilter instance
            The BilinearFilter instance.
        """
        if not isinstance(self.filters, np.ndarray):
            raise TypeError('filters must be an array.')
        if not isinstance(self.log, bool):
            raise TypeError('log must be a boolean')
        self.filters_ = self.filters
        return self

    def transform(self, X):
        """Apply spatial filters.

        Parameters
        ----------
        X : ndarray, shape (n_trials, n_channels, n_channels)
            Set of covariance matrices.

        Returns
        -------
        Xf : ndarray, shape (n_trials, n_filters) or \
                ndarray, shape (n_trials, n_filters, n_filters)
            Set of spatialy filtered log-variance or covariance, depending on
            the 'log' input parameter.
        """
        if not isinstance(X, (np.ndarray, list)):
            raise TypeError('X must be an array.')
        if X[0].shape[1] != self.filters_.shape[1]:
            raise ValueError("Data and filters dimension must be compatible.")

        X_filt = self.filters_ @ X @ self.filters_.T

        # if logvariance
        if self.log:
            out = np.zeros(X_filt.shape[:2])
            for i, x in enumerate(X_filt):
                out[i] = np.log(np.diag(x))
            return out
        else:
            return X_filt


class CSP(BilinearFilter):
    """CSP spatial filtering with covariance matrices as inputs.

    Implementation of the famous Common Spatial Pattern algorithm [1]_ [2]_,
    but with covariance matrices as input. In addition, the implementation
    allows different metric for the estimation of the class-related mean
    covariance matrices, as described in [3]_.

    This implementation support multiclass CSP by means of approximate joint
    diagonalization. In this case, the spatial filter selection is achieved
    according to [4]_.

    Parameters
    ----------
    nfilter : int, default=4
        The number of components to decompose M/EEG signals.
    metric : str, default="euclid"
        Metric used for the estimation of mean covariance matrices.
        For the list of supported metrics,
        see :func:`pyriemann.utils.mean.mean_covariance`.
    log : bool, default=True
        If true, return the log variance, otherwise return the spatially
        filtered covariance matrices.

    Attributes
    ----------
    filters_ : ndarray, shape (min(n_channels, n_filters), n_channels)
        If fit, the CSP spatial filters.
    patterns_ : ndarray, shape (min(n_channels, n_filters), n_channels)
        If fit, the CSP spatial patterns.

    See Also
    --------
    MDM, SPoC

    References
    ----------
    .. [1] `Spatial Patterns Underlying Population Differences in the
        Background EEG
        <https://link.springer.com/article/10.1007/BF01129656>`_
        Z. Koles, M. Lazar, and S. Zhou. Brain Topography 2(4), 275-284, 1990.
    .. [2] `Optimizing Spatial Filters for Robust EEG Single-Trial Analysis
        <https://ieeexplore.ieee.org/document/4408441>`_
        B. Blankertz, R. Tomioka, S. Lemm, M. Kawanabe, K-R. Muller. IEEE
        Signal Processing Magazine 25(1), 41-56, 2008.
    .. [3] `Common Spatial Pattern revisited by Riemannian geometry
        <https://hal.archives-ouvertes.fr/hal-00602686>`_
        A. Barachant, S. Bonnet, M. Congedo and C. Jutten. IEEE International
        Workshop on Multimedia Signal Processing (MMSP), p. 472-476, 2010.
    .. [4] `Multiclass common spatial patterns and information theoretic
        feature extraction
        <https://ieeexplore.ieee.org/document/4473042>`_
        IEEE Transactions on Biomedical Engineering, Volume 55, Issue 8,
        August 2008. pp. 1991 - 2000
    """

    def __init__(self, nfilter=4, metric="euclid", log=True):
        """Init."""
        self.nfilter = nfilter
        self.metric = metric
        self.log = log

    def fit(self, X, y):
        """Train CSP spatial filters.

        Parameters
        ----------
        X : ndarray, shape (n_trials, n_channels, n_channels)
            Set of covariance matrices.
        y : ndarray, shape (n_trials,)
            Labels for each trial.

        Returns
        -------
        self : CSP instance
            The CSP instance.
        """
        if not isinstance(self.nfilter, int):
            raise TypeError('nfilter must be an integer')
        if not isinstance(self.log, bool):
            raise TypeError('log must be a boolean')

        if not isinstance(X, (np.ndarray, list)):
            raise TypeError('X must be an array.')
        if not isinstance(y, (np.ndarray, list)):
            raise TypeError('y must be an array.')
        X, y = np.asarray(X), np.asarray(y)
        if X.ndim != 3:
            raise ValueError('X must be n_trials * n_channels * n_channels')
        if len(y) != len(X):
            raise ValueError('X and y must have the same length.')
        if np.squeeze(y).ndim != 1:
            raise ValueError('y must be of shape (n_trials,).')

        n_trials, n_channels, _ = X.shape
        classes = np.unique(y)
        # estimate class means
        C = []
        for c in classes:
            C.append(mean_covariance(X[y == c], self.metric))
        C = np.array(C)

        # Switch between binary and multiclass
        if len(classes) == 2:
            evals, evecs = eigh(C[1], C[0] + C[1])
            # sort eigenvectors
            ix = np.argsort(np.abs(evals - 0.5))[::-1]
        elif len(classes) > 2:
            evecs, D = ajd_pham(C)
            Ctot = mean_covariance(C, self.metric)
            evecs = evecs.T

            # normalize
            for i in range(evecs.shape[1]):
                tmp = evecs[:, i].T @ Ctot @ evecs[:, i]
                evecs[:, i] /= np.sqrt(tmp)

            mutual_info = []
            # class probability
            Pc = [np.mean(y == c) for c in classes]
            for j in range(evecs.shape[1]):
                a = 0
                b = 0
                for i, c in enumerate(classes):
                    tmp = evecs[:, j].T @ C[i] @ evecs[:, j]
                    a += Pc[i] * np.log(np.sqrt(tmp))
                    b += Pc[i] * (tmp ** 2 - 1)
                mi = - (a + (3.0 / 16) * (b ** 2))
                mutual_info.append(mi)
            ix = np.argsort(mutual_info)[::-1]
        else:
            raise ValueError("Number of classes must be >= 2.")

        # sort eigenvectors
        evecs = evecs[:, ix]

        # spatial patterns
        A = np.linalg.pinv(evecs.T)

        self.filters_ = evecs[:, 0:self.nfilter].T
        self.patterns_ = A[:, 0:self.nfilter].T

        return self


class SPoC(CSP):
    """SPoC spatial filtering with covariance matrices as inputs.

    Source Power Comodulation (SPoC) [1]_ allows to extract spatial filters and
    patterns by using a target (continuous) variable in the decomposition
    process in order to give preference to components whose power comodulates
    with the target variable.

    SPoC can be seen as an extension of the
    :class:`pyriemann.spatialfilters.CSP` driven by a continuous
    variable rather than a discrete (often binary) variable. Typical
    applications include extraction of motor patterns using EMG power or audio
    paterns using sound envelope.

    Parameters
    ----------
    nfilter : int, default=4
        The number of components to decompose M/EEG signals.
    metric : str, default="euclid"
        Metric used for the estimation of mean covariance matrices.
        For the list of supported metrics,
        see :func:`pyriemann.utils.mean.mean_covariance`.
    log : bool, default=True
        If true, return the log variance, otherwise return the spatially
        filtered covariance matrices.

    Attributes
    ----------
    filters_ : ndarray, shape (min(n_channels, n_filters), n_channels)
        If fit, the SPoC spatial filters.
    patterns_ : ndarray, shape (min(n_channels, n_filters), n_channels)
        If fit, the SPoC spatial patterns.

    Notes
    -----
    .. versionadded:: 0.2.4

    See Also
    --------
    CSP

    References
    ----------
    .. [1] `SPoC: a novel framework for relating the amplitude of neuronal
        oscillations to behaviorally relevant parameters
        <https://www.sciencedirect.com/science/article/pii/S1053811913008483>`_
        S. Dahne, F. C. Meinecke, S. Haufe, J. Hohne, M. Tangermann, K-R.
        Muller, and V. V. Nikulin. NeuroImage, 86, 111-122, 2014.
    """

    def fit(self, X, y):
        """Train spatial filters.

        Parameters
        ----------
        X : ndarray, shape (n_trials, n_channels, n_channels)
            Set of covariance matrices.
        y : ndarray, shape (n_trials,)
            Target variable for each trial.

        Returns
        -------
        self : SPoC instance
            The SPoC instance.
        """

        # Normalize target variable
        target = np.float64(y.copy())
        target -= target.mean()
        target /= target.std()

        C = mean_covariance(X, self.metric)
        Ce = np.zeros_like(X)
        for i in range(Ce.shape[0]):
            Ce[i] = X[i] * target[i]
        Cz = mean_covariance(Ce, self.metric)

        # solve eigenvalue decomposition
        evals, evecs = eigh(Cz, C)
        evals = evals.real
        evecs = evecs.real
        # sort vectors
        ix = np.argsort(np.abs(evals))[::-1]

        # sort eigenvectors
        evecs = evecs[:, ix]

        # spatial patterns
        A = np.linalg.pinv(evecs.T)

        self.filters_ = evecs[:, 0:self.nfilter].T
        self.patterns_ = A[:, 0:self.nfilter].T

        return self


class AJDC(BaseEstimator, TransformerMixin):
    """AJDC algorithm.

    The approximate joint diagonalization of Fourier cospectral matrices (AJDC)
    [1]_ is a versatile tool for blind source separation (BSS) tasks based on
    Second-Order Statistics (SOS), estimating spectrally uncorrelated sources.

    It can be applied:

    * on a single subject, to solve the classical BSS problem [1]_,
    * on several subjects, to solve the group BSS (gBSS) problem [2]_,
    * on several experimental conditions (for eg, baseline versus task), to
      exploit the diversity of source energy between conditions in addition
      to generic coloration and time-varying energy [1]_.

    AJDC estimates Fourier cospectral matrices by the Welch's method, and
    applies a trace-normalization. If necessary, it averages cospectra across
    subjects, and concatenates them along experimental conditions.
    Then, a dimension reduction and a whitening are applied on cospectra.
    An approximate joint diagonalization (AJD) [3]_ allows to estimate the
    joint diagonalizer, not constrained to be orthogonal. Finally, forward and
    backward spatial filters are computed.

    Parameters
    ----------
    window : int, default=128
        The length of the FFT window used for spectral estimation.
    overlap : float, default=0.5
        The percentage of overlap between window.
    fmin : float | None, default=None
        The minimal frequency to be returned. Since BSS models assume zero-mean
        processes, the first cospectrum (0 Hz) must be excluded.
    fmax : float | None, default=None
        The maximal frequency to be returned.
    fs : float | None, default=None
        The sampling frequency of the signal.
    dim_red : None | dict, default=None
        Parameter for dimension reduction of cospectra, because Pham's AJD is
        sensitive to matrices conditioning.

        If ``None`` :
            no dimension reduction during whitening.
        If ``{'n_components': val}`` :
            dimension reduction defining the number of components;
            ``val`` must be an integer superior to 1.
        If ``{'expl_var': val}`` :
            dimension reduction selecting the number of components such that
            the amount of variance that needs to be explained is greater than
            the percentage specified by ``val``.
            ``val`` must be a float in (0,1], typically ``0.99``.
        If ``{'max_cond': val}`` :
            dimension reduction selecting the number of components such that
            the condition number of the mean matrix is lower than ``val``.
            This threshold has a physiological interpretation, because it can
            be viewed as the ratio between the power of the strongest component
            (usually, eye-blink source) and the power of the lowest component
            you don't want to keep (acquisition sensor noise).
            ``val`` must be a float strictly superior to 1, typically 100.
        If ``{'warm_restart': val}`` :
            dimension reduction defining the number of components from an
            initial joint diagonalizer, and then run AJD from this solution.
            ``val`` must be a square ndarray.
    verbose : bool, default=True
        Verbose flag.

    Attributes
    ----------
    n_channels_ : int
        If fit, the number of channels of the signal.
    freqs_ : ndarray, shape (n_freqs,)
        If fit, the frequencies associated to cospectra.
    n_sources_ : int
        If fit, the number of components of the source space.
    diag_filters_ : ndarray, shape ``(n_sources_, n_sources_)``
        If fit, the diagonalization filters, also called joint diagonalizer.
    forward_filters_ : ndarray, shape ``(n_sources_, n_channels_)``
        If fit, the spatial filters used to transform signal into source,
        also called deximing or separating matrix.
    backward_filters_ : ndarray, shape ``(n_channels_, n_sources_)``
        If fit, the spatial filters used to transform source into signal,
        also called mixing matrix.

    Notes
    -----
    .. versionadded:: 0.2.7

    See Also
    --------
    CospCovariances

    References
    ----------
    .. [1] `On the blind source separation of human electroencephalogram by
        approximate joint diagonalization of second order statistics
        <https://hal.archives-ouvertes.fr/hal-00343628>`_
        M. Congedo, C. Gouy-Pailler, C. Jutten. Clinical Neurophysiology,
        Elsevier, 2008, 119 (12), pp.2677-2686.
    .. [2] `Group indepedent component analysis of resting state EEG in large
        normative samples
        <https://hal.archives-ouvertes.fr/hal-00523200>`_
        M. Congedo, R. John, D. de Ridder, L. Prichep. International Journal of
        Psychophysiology, Elsevier, 2010, 78, pp.89-99.
    .. [3] `Joint approximate diagonalization of positive definite
        Hermitian matrices
        <https://epubs.siam.org/doi/10.1137/S089547980035689X>`_
        D.-T. Pham. SIAM Journal on Matrix Analysis and Applications, Volume 22
        Issue 4, 2000
    """

    def __init__(self, window=128, overlap=0.5, fmin=None, fmax=None, fs=None,
                 dim_red=None, verbose=True):
        """Init."""

        self.window = window
        self.overlap = overlap
        self.fmin = fmin
        self.fmax = fmax
        self.fs = fs
        self.dim_red = dim_red
        self.verbose = verbose

    def fit(self, X, y=None):
        """Fit.

        Compute and diagonalize cospectra, to estimate forward and backward
        spatial filters.

        Parameters
        ----------
        X : ndarray, shape (n_subjects, n_conditions, n_channels, n_times) | \
                list of n_subjects of list of n_conditions ndarray of shape \
                (n_channels, n_times), with same n_conditions and n_channels \
                but different n_times
            Multi-channel time-series in channel space, acquired for different
            subjects and under different experimental conditions.
        y : None
            Currently not used, here for compatibility with sklearn API.

        Returns
        -------
        self : AJDC instance
            The AJDC instance.
        """
        # definition of params for Welch's method
        cospcov = est.CospCovariances(
            window=self.window,
            overlap=self.overlap,
            fmin=self.fmin,
            fmax=self.fmax,
            fs=self.fs)
        # estimation of cospectra on subjects and conditions
        cosp = []
        for s in range(len(X)):
            cosp_ = cospcov.transform(X[s])
            if s == 0:
                n_conditions = cosp_.shape[0]
                self.n_channels_ = cosp_.shape[1]
                self.freqs_ = cospcov.freqs_
            else:
                if n_conditions != cosp_.shape[0]:
                    raise ValueError('Unequal number of conditions')
                if self.n_channels_ != cosp_.shape[1]:
                    raise ValueError('Unequal number of channels')
            cosp.append(cosp_)
        cosp = np.transpose(np.array(cosp), axes=(0, 1, 4, 2, 3))

        # trace-normalization of cospectra, Eq(3) in [2]
        cosp = normalize(cosp, "trace")
        # average of cospectra across subjects, Eq(7) in [2]
        cosp = np.mean(cosp, axis=0, keepdims=False)
        # concatenation of cospectra along conditions
        self._cosp_channels = np.concatenate(cosp, axis=0)
        # estimation of non-diagonality weights, Eq(B.1) in [1]
        weights = get_nondiag_weight(self._cosp_channels)

        # initial diagonalizer: if warm restart, dimension reduction defined by
        # the size of the initial diag filters
        init = None
        if self.dim_red is None:
            warnings.warn('Parameter dim_red should not be let to None')
        elif isinstance(self.dim_red, dict) and len(self.dim_red) == 1 \
                and next(iter(self.dim_red)) == 'warm_restart':
            init = self.dim_red['warm_restart']
            if init.ndim != 2 or init.shape[0] != init.shape[1]:
                raise ValueError(
                    'Initial diagonalizer defined in dim_red is not a 2D '
                    'square matrix (Got shape = %s).' % (init.shape,))
            self.dim_red = {'n_components': init.shape[0]}

        # dimension reduction and whitening, Eq.(8) in [2], computed on the
        # weighted mean of cospectra across frequencies (and conditions)
        whit = Whitening(
            metric='euclid',
            dim_red=self.dim_red,
            verbose=self.verbose)
        cosp_rw = whit.fit_transform(self._cosp_channels, weights)
        self.n_sources_ = whit.n_components_

        # approximate joint diagonalization, currently by Pham's algorithm [3]
        self.diag_filters_, self._cosp_sources = ajd_pham(
            cosp_rw,
            init=init,
            n_iter_max=100,
            sample_weight=weights)

        # computation of forward and backward filters, Eq.(9) and (10) in [2]
        self.forward_filters_ = self.diag_filters_ @ whit.filters_.T
        self.backward_filters_ = whit.inv_filters_.T @ inv(self.diag_filters_)
        return self

    def transform(self, X):
        """Transform channel space to source space.

        Transform channel space to source space, applying forward spatial
        filters.

        Parameters
        ----------
        X : ndarray, shape (n_matrices, n_channels, n_times)
            Multi-channel time-series in channel space.

        Returns
        -------
        source : ndarray, shape (n_matrices, n_sources, n_times)
            Multi-channel time-series in source space.
        """
        if X.ndim != 3:
            raise ValueError('X must have 3 dimensions (Got %d)' % X.ndim)
        if X.shape[1] != self.n_channels_:
            raise ValueError(
                'X does not have the good number of channels. Should be %d but'
                ' got %d.' % (self.n_channels_, X.shape[1]))

        source = self.forward_filters_ @ X
        return source

    def inverse_transform(self, X, supp=None):
        """Transform source space to channel space.

        Transform source space to channel space, applying backward spatial
        filters, with the possibility to suppress some sources, like in BSS
        filtering/denoising.

        Parameters
        ----------
        X : ndarray, shape (n_matrices, n_sources, n_times)
            Multi-channel time-series in source space.
        supp : list of int | None, default=None
            Indices of sources to suppress. If None, no source suppression.

        Returns
        -------
        signal : ndarray, shape (n_matrices, n_channels, n_times)
            Multi-channel time-series in channel space.
        """
        if X.ndim != 3:
            raise ValueError('X must have 3 dimensions (Got %d)' % X.ndim)
        if X.shape[1] != self.n_sources_:
            raise ValueError(
                'X does not have the good number of sources. Should be %d but '
                'got %d.' % (self.n_sources_, X.shape[1]))

        denois = np.eye(self.n_sources_)
        if supp is None:
            pass
        elif isinstance(supp, list):
            for s in supp:
                denois[s, s] = 0
        else:
            raise ValueError('Parameter supp must be a list of int, or None')

        signal = self.backward_filters_ @ denois @ X
        return signal

    def get_src_expl_var(self, X):
        """Estimate explained variances of sources.

        Estimate explained variances of sources, see Appendix D in [1].

        Parameters
        ----------
        X : ndarray, shape (n_matrices, n_channels, n_times)
            Multi-channel time-series in channel space.

        Returns
        -------
        src_var : ndarray, shape (n_matrices, n_sources)
            Explained variance for each source.
        """
        if X.ndim != 3:
            raise ValueError('X must have 3 dimensions (Got %d)' % X.ndim)
        if X.shape[1] != self.n_channels_:
            raise ValueError(
                'X does not have the good number of channels. Should be %d but'
                ' got %d.' % (self.n_channels_, X.shape[1]))

        cov = est.Covariances().transform(X)

        src_var = np.zeros((X.shape[0], self.n_sources_))
        for s in range(self.n_sources_):
            src_var[:, s] = np.trace(
                self.backward_filters_[:, s] * self.forward_filters_[s].T * cov
                * self.forward_filters_[s] * self.backward_filters_[:, s].T,
                axis1=-2,
                axis2=-1)
        return src_var



"""Distances between SPD/HPD matrices."""

import numpy as np
from scipy.linalg import eigvalsh, solve
from sklearn.metrics import euclidean_distances

from .base import logm, sqrtm, invsqrtm
from .utils import check_function


def _check_inputs(A, B):
    if not isinstance(A, np.ndarray) or not isinstance(B, np.ndarray):
        raise ValueError("Inputs must be ndarrays")
    if not A.shape == B.shape:
        raise ValueError("Inputs must have equal dimensions")
    if A.ndim < 2:
        raise ValueError("Inputs must be at least a 2D ndarray")


def _recursive(fun, A, B, *args, **kwargs):
    """Recursive function with two inputs."""
    if A.ndim == 2:
        return fun(A, B, *args, **kwargs)
    else:
        return np.asarray(
            [_recursive(fun, a, b, *args, **kwargs) for a, b in zip(A, B)]
        )


###############################################################################
# Distances between matrices


def distance_euclid(A, B, squared=False):
    r"""Euclidean distance between matrices.

    The Euclidean distance between two matrices :math:`\mathbf{A}` and
    :math:`\mathbf{B}` is defined as the Frobenius norm of the difference of
    the two matrices:

    .. math::
        d(\mathbf{A},\mathbf{B}) = \Vert \mathbf{A} - \mathbf{B} \Vert_F

    Parameters
    ----------
    A : ndarray, shape (..., n, m)
        First matrices, at least 2D ndarray.
    B : ndarray, shape (..., n, m)
        Second matrices, same dimensions as A.
    squared : bool, default False
        Return squared distance.

        .. versionadded:: 0.5

    Returns
    -------
    d : float or ndarray, shape (...,)
        Euclidean distance between A and B.

    See Also
    --------
    distance
    """
    _check_inputs(A, B)
    d = np.linalg.norm(A - B, ord='fro', axis=(-2, -1))
    return d ** 2 if squared else d


def distance_harmonic(A, B, squared=False):
    r"""Harmonic distance between invertible matrices.

    The harmonic distance between two invertible matrices :math:`\mathbf{A}`
    and :math:`\mathbf{B}` is:

    .. math::
        d(\mathbf{A},\mathbf{B}) =
        \Vert \mathbf{A}^{-1} - \mathbf{B}^{-1} \Vert_F

    Parameters
    ----------
    A : ndarray, shape (..., n, n)
        First invertible matrices, at least 2D ndarray.
    B : ndarray, shape (..., n, n)
        Second invertible matrices, same dimensions as A.
    squared : bool, default False
        Return squared distance.

        .. versionadded:: 0.5

    Returns
    -------
    d : float or ndarray, shape (...,)
        Harmonic distance between A and B.

    See Also
    --------
    distance
    """
    return distance_euclid(np.linalg.inv(A), np.linalg.inv(B), squared=squared)


def distance_kullback(A, B, squared=False):
    r"""Kullback-Leibler divergence between SPD/HPD matrices.

    The left Kullback-Leibler divergence between two SPD/HPD matrices
    :math:`\mathbf{A}` and :math:`\mathbf{B}` is [1]_:

    .. math::
        d(\mathbf{A},\mathbf{B}) =
        \frac{1}{2} \left( \text{tr}(\mathbf{B}^{-1}\mathbf{A}) - n
        + \log \left( \frac{\det(\mathbf{B})}{\det(\mathbf{A})}\right) \right)

    Parameters
    ----------
    A : ndarray, shape (..., n, n)
        First SPD/HPD matrices, at least 2D ndarray.
    B : ndarray, shape (..., n, n)
        Second SPD/HPD matrices, same dimensions as A.
    squared : bool, default False
        Return squared distance.

        .. versionadded:: 0.5

    Returns
    -------
    d : float or ndarray, shape (...,)
        Left Kullback-Leibler divergence between A and B.

    See Also
    --------
    distance

    References
    ----------
    .. [1] `On information and sufficiency
        <https://www.jstor.org/stable/2236703>`_
        S. Kullback S, R. Leibler.
        The Annals of Mathematical Statistics, 1951, 22 (1), pp. 79-86
    """
    _check_inputs(A, B)
    n = A.shape[-1]
    tr = np.trace(_recursive(solve, B, A, assume_a='pos'), axis1=-2, axis2=-1)
    logdet = np.linalg.slogdet(B)[1] - np.linalg.slogdet(A)[1]
    d = 0.5 * (tr - n + logdet).real
    return d ** 2 if squared else d


def distance_kullback_right(A, B, squared=False):
    """Wrapper for right Kullback-Leibler divergence."""
    return distance_kullback(B, A, squared=squared)


def distance_kullback_sym(A, B, squared=False):
    r"""Symmetrized Kullback-Leibler divergence between SPD/HPD matrices.

    The symmetrized Kullback-Leibler divergence between two SPD/HPD matrices
    :math:`\mathbf{A}` and :math:`\mathbf{B}` is the sum of left and right
    Kullback-Leibler divergences.
    It is also called Jeffreys divergence [1]_.

    Parameters
    ----------
    A : ndarray, shape (..., n, n)
        First SPD/HPD matrices, at least 2D ndarray.
    B : ndarray, shape (..., n, n)
        Second SPD/HPD matrices, same dimensions as A.
    squared : bool, default False
        Return squared distance.

        .. versionadded:: 0.5

    Returns
    -------
    d : float or ndarray, shape (...,)
        Symmetrized Kullback-Leibler divergence between A and B.

    See Also
    --------
    distance

    References
    ----------
    .. [1] `An invariant form for the prior probability in estimation problems
        <https://www.jstor.org/stable/97883>`_
        H. Jeffreys.
        Proceedings of the Royal Society of London A: mathematical, physical
        and engineering sciences, 1946, 186 (1007), pp. 453-461
    """
    d = distance_kullback(A, B) + distance_kullback_right(A, B)
    return d ** 2 if squared else d


def distance_logdet(A, B, squared=False):
    r"""Log-det distance between SPD/HPD matrices.

    The log-det distance between two SPD/HPD matrices :math:`\mathbf{A}` and
    :math:`\mathbf{B}` is [1]_:

    .. math::
        d(\mathbf{A},\mathbf{B}) =
        \sqrt{\log(\det \left( \frac{\mathbf{A}+\mathbf{B}}{2} \right))
        - \frac{1}{2} \log(\det(\mathbf{A} \mathbf{B}))}

    Parameters
    ----------
    A : ndarray, shape (..., n, n)
        First SPD/HPD matrices, at least 2D ndarray.
    B : ndarray, shape (..., n, n)
        Second SPD/HPD matrices, same dimensions as A.
    squared : bool, default False
        Return squared distance.

        .. versionadded:: 0.5

    Returns
    -------
    d : float or ndarray, shape (...,)
        Log-det distance between A and B.

    See Also
    --------
    distance

    References
    ----------
    .. [1] `Matrix nearness problems with Bregman divergences
        <https://epubs.siam.org/doi/abs/10.1137/060649021>`_
        I.S. Dhillon, J.A. Tropp.
        SIAM J Matrix Anal Appl, 2007, 29 (4), pp. 1120-1146
    """
    _check_inputs(A, B)
    logdet_ApB = np.linalg.slogdet((A + B) / 2.0)[1]
    logdet_AxB = np.linalg.slogdet(A @ B)[1]
    d2 = logdet_ApB - 0.5 * logdet_AxB
    d2 = np.maximum(0, d2)
    return d2 if squared else np.sqrt(d2)


def distance_logeuclid(A, B, squared=False):
    r"""Log-Euclidean distance between SPD/HPD matrices.

    The Log-Euclidean distance between two SPD/HPD matrices :math:`\mathbf{A}`
    and :math:`\mathbf{B}` is [1]_:

    .. math::
        d(\mathbf{A},\mathbf{B}) =
        \Vert \log(\mathbf{A}) - \log(\mathbf{B}) \Vert_F

    Parameters
    ----------
    A : ndarray, shape (..., n, n)
        First SPD/HPD matrices, at least 2D ndarray.
    B : ndarray, shape (..., n, n)
        Second SPD/HPD matrices, same dimensions as A.
    squared : bool, default False
        Return squared distance.

        .. versionadded:: 0.5

    Returns
    -------
    d : float or ndarray, shape (...,)
        Log-Euclidean distance between A and B.

    See Also
    --------
    distance

    References
    ----------
    .. [1] `Geometric means in a novel vector space structure on symmetric
        positive-definite matrices
        <https://epubs.siam.org/doi/abs/10.1137/050637996>`_
        V. Arsigny, P. Fillard, X. Pennec, N. Ayache.
        SIAM J Matrix Anal Appl, 2007, 29 (1), pp. 328-347
    """
    return distance_euclid(logm(A), logm(B), squared=squared)


def distance_riemann(A, B, squared=False):
    r"""Affine-invariant Riemannian distance between SPD/HPD matrices.

    The affine-invariant Riemannian distance between two SPD/HPD matrices
    :math:`\mathbf{A}` and :math:`\mathbf{B}` is [1]_:

    .. math::
        d(\mathbf{A},\mathbf{B}) =
        {\left( \sum_i \log(\lambda_i)^2 \right)}^{1/2}

    where :math:`\lambda_i` are the joint eigenvalues of :math:`\mathbf{A}` and
    :math:`\mathbf{B}`.

    Parameters
    ----------
    A : ndarray, shape (..., n, n)
        First SPD/HPD matrices, at least 2D ndarray.
    B : ndarray, shape (..., n, n)
        Second SPD/HPD matrices, same dimensions as A.
    squared : bool, default False
        Return squared distance.

        .. versionadded:: 0.5

    Returns
    -------
    d : float or ndarray, shape (...,)
        Affine-invariant Riemannian distance between A and B.

    See Also
    --------
    distance

    References
    ----------
    .. [1] `A differential geometric approach to the geometric mean of
        symmetric positive-definite matrices
        <https://epubs.siam.org/doi/10.1137/S0895479803436937>`_
        M. Moakher. SIAM J Matrix Anal Appl, 2005, 26 (3), pp. 735-747
    """
    _check_inputs(A, B)
    d2 = (np.log(_recursive(eigvalsh, A, B))**2).sum(axis=-1)
    return d2 if squared else np.sqrt(d2)


def distance_wasserstein(A, B, squared=False):
    r"""Wasserstein distance between SPSD/HPSD matrices.

    The Wasserstein distance between two SPSD/HPSD matrices :math:`\mathbf{A}`
    and :math:`\mathbf{B}` is [1]_ [2]_:

    .. math::
        d(\mathbf{A},\mathbf{B}) =
        \sqrt{ \text{tr} \left(\mathbf{A} + \mathbf{B}
        - 2(\mathbf{B}^{1/2} \mathbf{A} \mathbf{B}^{1/2})^{1/2} \right) }

    Parameters
    ----------
    A : ndarray, shape (..., n, n)
        First SPSD/HPSD matrices, at least 2D ndarray.
    B : ndarray, shape (..., n, n)
        Second SPSD/HPSD matrices, same dimensions as A.
    squared : bool, default False
        Return squared distance.

        .. versionadded:: 0.5

    Returns
    -------
    d : float or ndarray, shape (...,)
        Wasserstein distance between A and B.

    See Also
    --------
    distance

    References
    ----------
    .. [1] `Optimal transport: old and new
        <https://link.springer.com/book/10.1007/978-3-540-71050-9>`_
        C. Villani. Springer Science & Business Media, 2008, vol. 338
    .. [2] `An extension of Kakutani's theorem on infinite product measures to
        the tensor product of semifinite w*-algebras
        <https://www.ams.org/journals/tran/1969-135-00/S0002-9947-1969-0236719-2/S0002-9947-1969-0236719-2.pdf>`_
        D. Bures. Trans Am Math Soc, 1969, 135, pp. 199-212
    """  # noqa
    _check_inputs(A, B)
    B12 = sqrtm(B)
    d2 = np.trace(A + B - 2 * sqrtm(B12 @ A @ B12), axis1=-2, axis2=-1)
    d2 = np.maximum(0, d2.real)
    return d2 if squared else np.sqrt(d2)


distance_functions = {
    "euclid": distance_euclid,
    "harmonic": distance_harmonic,
    "kullback": distance_kullback,
    "kullback_right": distance_kullback_right,
    "kullback_sym": distance_kullback_sym,
    "logdet": distance_logdet,
    "logeuclid": distance_logeuclid,
    "riemann": distance_riemann,
    "wasserstein": distance_wasserstein,
}


def distance(A, B, metric="riemann", squared=False):
    """Distance between matrices according to a metric.

    Compute the distance between two matrices A and B according to a metric
    [1]_, or between a set of matrices A and another matrix B.

    Parameters
    ----------
    A : ndarray, shape (n, n) or shape (n_matrices, n, n)
        First matrix, or set of matrices.
    B : ndarray, shape (n, n)
        Second matrix.
    metric : string | callable, default="riemann"
        Metric for distance, can be: "euclid", "harmonic", "kullback",
        "kullback_right", "kullback_sym", "logdet", "logeuclid", "riemann",
        "wasserstein", or a callable function.
    squared : bool, default False
        Return squared distance.

        .. versionadded:: 0.5

    Returns
    -------
    d : float or ndarray, shape (n_matrices, 1)
        Distance between A and B.

    References
    ----------
    .. [1] `Review of Riemannian distances and divergences, applied to
        SSVEP-based BCI
        <https://hal.archives-ouvertes.fr/LISV/hal-03015762v1>`_
        S. Chevallier, E. K. Kalunga, Q. Barthélemy, E. Monacelli.
        Neuroinformatics, Springer, 2021, 19 (1), pp.93-106
    """
    distance_function = check_function(metric, distance_functions)

    shape_A, shape_B = A.shape, B.shape
    if shape_A == shape_B:
        d = distance_function(A, B, squared=squared)
    elif len(shape_A) == 3 and len(shape_B) == 2:
        d = np.empty((shape_A[0], 1))
        for i in range(shape_A[0]):
            d[i] = distance_function(A[i], B, squared=squared)
    else:
        raise ValueError("Inputs have incompatible dimensions.")

    return d


###############################################################################


def _pairwise_distance_euclid(X, Y=None, squared=False):
    """Pairwise Euclidean distance matrix.

    Compute the matrix of Euclidean distances between pairs of elements of X
    and Y.

    Parameters
    ----------
    X : ndarray, shape (n_matrices_X, n, n)
        First set of matrices.
    Y : None | ndarray, shape (n_matrices_Y, n, n), default=None
        Second set of matrices. If None, Y is set to X.
    squared : bool, default False
        Return squared distances.

    Returns
    -------
    dist : ndarray, shape (n_matrices_X, n_matrices_X) or (n_matrices_X, \
            n_matrices_Y)
        Euclidean distances between pairs of elements of X if Y is None, or
        between elements of X and Y.

    See Also
    --------
    pairwise_distance
    distance_euclid
    """
    if Y is None:
        dist = euclidean_distances(X.reshape(len(X), -1), squared=squared)
    else:
        dist = euclidean_distances(X.reshape(len(X), -1),
                                   Y.reshape(len(Y), -1),
                                   squared=squared)
    return dist


def _pairwise_distance_harmonic(X, Y=None, squared=False):
    """Pairwise harmonic distance matrix.

    Compute the matrix of harmonic distances between pairs of elements of X and
    Y.

    Parameters
    ----------
    X : ndarray, shape (n_matrices_X, n, n)
        First set of matrices.
    Y : None | ndarray, shape (n_matrices_Y, n, n), default=None
        Second set of matrices. If None, Y is set to X.
    squared : bool, default False
        Return squared distances.

    Returns
    -------
    dist : ndarray, shape (n_matrices_X, n_matrices_X) or (n_matrices_X, \
            n_matrices_Y)
        Harmonic distances between pairs of elements of X if Y is None, or
        between elements of X and Y.

    See Also
    --------
    pairwise_distance
    distance_harmonic
    """
    invX = np.linalg.inv(X)
    if Y is None:
        invY = None
    else:
        invY = np.linalg.inv(Y)

    return _pairwise_distance_euclid(invX, invY, squared=squared)


def _pairwise_distance_logeuclid(X, Y=None, squared=False):
    """Pairwise Log-Euclidean distance matrix.

    Compute the matrix of Log-Euclidean distances between pairs of elements of
    X and Y.

    Parameters
    ----------
    X : ndarray, shape (n_matrices_X, n, n)
        First set of matrices.
    Y : None | ndarray, shape (n_matrices_Y, n, n), default=None
        Second set of matrices. If None, Y is set to X.
    squared : bool, default False
        Return squared distances.

    Returns
    -------
    dist : ndarray, shape (n_matrices_X, n_matrices_X) or (n_matrices_X, \
            n_matrices_Y)
        Log-Euclidean distances between pairs of elements of X if Y is None, or
        between elements of X and Y.

    See Also
    --------
    pairwise_distance
    distance_logeuclid
    """
    logX = logm(X)
    if Y is None:
        logY = None
    else:
        logY = logm(Y)

    return _pairwise_distance_euclid(logX, logY, squared=squared)


def _pairwise_distance_riemann(X, Y=None, squared=False):
    """Pairwise Riemannian distance matrix.

    Compute the matrix of Riemannian distances between pairs of elements of X
    and Y.

    Parameters
    ----------
    X : ndarray, shape (n_matrices_X, n, n)
        First set of matrices.
    Y : None | ndarray, shape (n_matrices_Y, n, n), default=None
        Second set of matrices. If None, Y is set to X.
    squared : bool, default False
        Return squared distances.

    Returns
    -------
    dist : ndarray, shape (n_matrices_X, n_matrices_X) or (n_matrices_X, \
            n_matrices_Y)
        Riemannian distances between pairs of elements of X if Y is None, or
        between elements of X and Y.

    See Also
    --------
    pairwise_distance
    distance_riemann
    """
    XisY = False
    if Y is None:
        XisY = True
        Y = X

    n_matrices_X, n_matrices_Y = len(X), len(Y)
    Xinv12 = invsqrtm(X)
    dist = np.zeros((n_matrices_X, n_matrices_Y))

    # row by row so it fits in memory
    for i, x_ in enumerate(Xinv12):
        evals_ = np.linalg.eigvalsh(x_ @ Y[i * XisY:] @ x_)
        d2 = np.sum(np.log(evals_) ** 2, -1)
        dist[i, i * XisY:] = d2

    if XisY:
        dist += dist.T

    return dist if squared else np.sqrt(dist)


def pairwise_distance(X, Y=None, metric="riemann", squared=False):
    """Pairwise distance matrix.

    Compute the matrix of distances between pairs of elements of X and Y.

    Parameters
    ----------
    X : ndarray, shape (n_matrices_X, n, n)
        First set of matrices.
    Y : None | ndarray, shape (n_matrices_Y, n, n), default=None
        Second set of matrices. If None, Y is set to X.
    metric : string, default='riemann'
        Metric for pairwise distance. For the list of supported metrics,
        see :func:`pyriemann.utils.distance.distance`.
    squared : bool, default False
        Return squared distances.

        .. versionadded:: 0.5

    Returns
    -------
    dist : ndarray, shape (n_matrices_X, n_matrices_X) or (n_matrices_X, \
            n_matrices_Y)
        Distances between pairs of elements of X if Y is None, or between
        elements of X and Y.

    See Also
    --------
    distance
    """
    if metric == "euclid":
        return _pairwise_distance_euclid(X, Y=Y, squared=squared)
    elif metric == "harmonic":
        return _pairwise_distance_harmonic(X, Y=Y, squared=squared)
    elif metric == "logeuclid":
        return _pairwise_distance_logeuclid(X, Y=Y, squared=squared)
    elif metric == "riemann":
        return _pairwise_distance_riemann(X, Y=Y, squared=squared)

    n_matrices_X, _, _ = X.shape

    # compute full pairwise matrix for non-symmetric metrics
    if Y is None and metric in ["kullback", "kullback_right"]:
        Y = X

    if Y is None:
        dist = np.zeros((n_matrices_X, n_matrices_X))
        for i in range(n_matrices_X):
            for j in range(i + 1, n_matrices_X):
                dist[i, j] = distance(X[i], X[j], metric, squared=squared)
        dist += dist.T
    else:
        n_matrices_Y, _, _ = Y.shape
        dist = np.empty((n_matrices_X, n_matrices_Y))
        for i in range(n_matrices_X):
            for j in range(n_matrices_Y):
                dist[i, j] = distance(X[i], Y[j], metric, squared=squared)

    return dist


###############################################################################
# Distances between vectors and matrices


def distance_mahalanobis(X, cov, mean=None, squared=False):
    r"""Mahalanobis distance between vectors and a Gaussian distribution.

    The Mahalanobis distance between a vector :math:`x \in \mathbb{C}^n` and a
    multivariate Gaussian distribution :math:`\mathcal{N}(\mu, C)`, with mean
    vector :math:`\mu \in \mathbb{C}^n` and covariance matrix
    :math:`C \in \mathbb{C}^{n \times n}` , is:

    .. math::
        d(x, \mathcal{N}(\mu, C)) = \sqrt{ (x - \mu)^H C^{-1} (x - \mu) }

    Parameters
    ----------
    X : ndarray, shape (n, n_vectors)
        Vectors.
    cov : ndarray, shape (n, n)
        Covariance matrix of the multivariate Gaussian distribution.
    mean : None | ndarray, shape (n, 1), default=None
        Mean vector of the multivariate Gaussian distribution.
        If None, distribution is considered as centered.
    squared : bool, default False
        Return squared distance.

        .. versionadded:: 0.5

    Returns
    -------
    d : ndarray, shape (n_vectors,)
        Mahalanobis distances.

    Notes
    -----
    .. versionadded:: 0.4

    References
    ----------
    .. [1] https://docs.scipy.org/doc/scipy/reference/generated/scipy.spatial.distance.mahalanobis.html
    """  # noqa
    if mean is not None:
        X -= mean

    Xw = invsqrtm(cov) @ X
    d2 = np.einsum('ij,ji->i', Xw.conj().T, Xw).real
    return d2 if squared else np.sqrt(d2)



from ._tools import (
    encode_domains,
    decode_domains,
    TLSplitter
)

from ._estimators import (
    TLDummy,
    TLCenter,
    TLStretch,
    TLRotate,
    TLEstimator,
    TLClassifier,
    TLRegressor,
    MDWM,
)

__all__ = [
    "encode_domains",
    "decode_domains",
    "TLDummy",
    "TLSplitter",
    "TLCenter",
    "TLStretch",
    "TLRotate",
    "TLEstimator",
    "TLClassifier",
    "TLRegressor",
    "MDWM",
]
